# -*- coding: utf-8 -*-
from ...compat import Apply, Op, tt
import numpy as np

__all__ = ["pTOp"]


class pTOp(Op):
    def __init__(self, func, deg):
        self.func = func
        self.deg = deg
        self.N = (deg + 1) ** 2
        self._grad_op = pTGradientOp(self)

    def make_node(self, *inputs):
        inputs = [tt.as_tensor_variable(i) for i in inputs]
        outputs = [tt.TensorType(inputs[0].dtype, (False, False))()]
        return Apply(self, inputs, outputs)

    def infer_shape(self, *args):
        shapes = args[-1]
        return [[shapes[0][0], self.N]]

    def perform(self, node, inputs, outputs):
        outputs[0][0] = self.func(self.deg, *inputs)

    def grad(self, inputs, gradients):
        return self._grad_op(*(inputs + gradients))


class pTGradientOp(Op):
    def __init__(self, base_op):
        self.base_op = base_op

        # Pre-compute the gradient factors for x, y, and z
        n = 0
        self.xf = np.zeros(self.base_op.N, dtype=int)
        self.yf = np.zeros(self.base_op.N, dtype=int)
        self.zf = np.zeros(self.base_op.N, dtype=int)
        for l in range(self.base_op.deg + 1):
            for m in range(-l, l + 1):
                mu = l - m
                nu = l + m
                if nu % 2 == 0:
                    if mu > 0:
                        self.xf[n] = mu // 2
                    if nu > 0:
                        self.yf[n] = nu // 2
                else:
                    if mu > 1:
                        self.xf[n] = (mu - 1) // 2
                    if nu > 1:
                        self.yf[n] = (nu - 1) // 2
                    self.zf[n] = 1
                n += 1

    def make_node(self, *inputs):
        inputs = [tt.as_tensor_variable(i) for i in inputs]
        outputs = [i.type() for i in inputs[:-1]]
        return Apply(self, inputs, outputs)

    def infer_shape(self, *args):
        shapes = args[-1]
        return shapes[:-1]

    def perform(self, node, inputs, outputs):
        x, y, z, bpT = inputs

        # TODO: When any of the coords are zero, there's a div
        # by zero below. This hack fixes the issue. We should
        # think of a better way of doing this!
        tol = 1e-8
        x[np.abs(x) < tol] = tol
        y[np.abs(y) < tol] = tol
        z[np.abs(z) < tol] = tol

        bpTpT = bpT * self.base_op.func(self.base_op.deg, x, y, z)
        bx = np.nansum(self.xf[None, :] * bpTpT / x[:, None], axis=-1)
        by = np.nansum(self.yf[None, :] * bpTpT / y[:, None], axis=-1)
        bz = np.nansum(self.zf[None, :] * bpTpT / z[:, None], axis=-1)
        outputs[0][0] = np.reshape(bx, np.shape(inputs[0]))
        outputs[1][0] = np.reshape(by, np.shape(inputs[1]))
        outputs[2][0] = np.reshape(bz, np.shape(inputs[2]))



# -*- coding: utf-8 -*-
from ...compat import Apply, Op, tt
import numpy as np

__all__ = ["spotYlmOp"]


class spotYlmOp(Op):
    def __init__(self, func, ydeg, nw):
        self.func = func
        self._grad_op = spotYlmGradientOp(self)
        self.Ny = (ydeg + 1) ** 2
        self.nw = nw

    def make_node(self, *inputs):
        inputs = [tt.as_tensor_variable(i) for i in inputs]
        if self.nw is None:
            outputs = [tt.TensorType(inputs[0].dtype, (False,))()]
        else:
            outputs = [tt.TensorType(inputs[0].dtype, (False, False))()]
        return Apply(self, inputs, outputs)

    def infer_shape(self, *args):
        if self.nw is None:
            return [(self.Ny,)]
        else:
            return [(self.Ny, self.nw)]

    def perform(self, node, inputs, outputs):
        outputs[0][0] = self.func(*inputs)
        if self.nw is None:
            outputs[0][0] = np.reshape(outputs[0][0], -1)

    def grad(self, inputs, gradients):
        return self._grad_op(*(inputs + gradients))


class spotYlmGradientOp(Op):
    def __init__(self, base_op):
        self.base_op = base_op

    def make_node(self, *inputs):
        inputs = [tt.as_tensor_variable(i) for i in inputs]
        outputs = [i.type() for i in inputs[:-1]]
        return Apply(self, inputs, outputs)

    def infer_shape(self, *args):
        shapes = args[-1]
        return shapes[:-1]

    def perform(self, node, inputs, outputs):
        bamp, bsigma, blat, blon = self.base_op.func(*inputs)
        outputs[0][0] = np.reshape(bamp, np.shape(inputs[0]))
        outputs[1][0] = np.reshape(bsigma, np.shape(inputs[1]))
        outputs[2][0] = np.reshape(blat, np.shape(inputs[2]))
        outputs[3][0] = np.reshape(blon, np.shape(inputs[3]))



# -*- coding: utf-8 -*-
from . import config
from ._constants import *
from .maps import MapBase, RVBase, ReflectedBase
from ._core import OpsSystem, math
from .compat import evaluator
import numpy as np
from astropy import units
from inspect import getmro
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
from IPython.display import HTML
import os
import logging

logger = logging.getLogger("starry.maps")


__all__ = ["Primary", "Secondary", "System"]


class Body(object):
    """A generic body. Must be subclassed."""

    def __init__(
        self,
        map,
        r=1.0,
        m=1.0,
        prot=1.0,
        t0=0.0,
        theta0=0.0,
        length_unit=units.Rsun,
        mass_unit=units.Msun,
        time_unit=units.day,
        angle_unit=units.degree,
        **kwargs,
    ):
        # Surface map
        self._lazy = map._lazy
        self._math = map._math
        self.map = map

        # Units
        self.length_unit = length_unit
        self.mass_unit = mass_unit
        self.time_unit = time_unit
        self.angle_unit = angle_unit

        # Attributes
        self.r = r
        self.m = m
        self.prot = prot
        self.t0 = t0
        self.theta0 = theta0

    @property
    def length_unit(self):
        """An ``astropy.units`` unit defining the length metric for this body."""
        return self._length_unit

    @length_unit.setter
    def length_unit(self, value):
        assert value.physical_type == "length"
        self._length_unit = value
        self._length_factor = value.in_units(units.Rsun)

    @property
    def mass_unit(self):
        """An ``astropy.units`` unit defining the mass metric for this body."""
        return self._mass_unit

    @mass_unit.setter
    def mass_unit(self, value):
        assert value.physical_type == "mass"
        self._mass_unit = value
        self._mass_factor = value.in_units(units.Msun)

    @property
    def time_unit(self):
        """An ``astropy.units`` unit defining the time metric for this body."""
        return self._time_unit

    @time_unit.setter
    def time_unit(self, value):
        assert value.physical_type == "time"
        self._time_unit = value
        self._time_factor = value.in_units(units.day)

    @property
    def angle_unit(self):
        """An ``astropy.units`` unit defining the angle metric for this body."""
        return self._angle_unit

    @angle_unit.setter
    def angle_unit(self, value):
        assert value.physical_type == "angle"
        self._angle_unit = value
        self._angle_factor = value.in_units(units.radian)

    @property
    def _angle_unit(self):
        return self._map._angle_unit

    @_angle_unit.setter
    def _angle_unit(self, value):
        self._map._angle_unit = value

    @property
    def _angle_factor(self):
        return self._map._angle_factor

    @_angle_factor.setter
    def _angle_factor(self, value):
        self._map._angle_factor = value

    @property
    def map(self):
        """The surface map for this body."""
        return self._map

    @map.setter
    def map(self, value):
        assert MapBase in getmro(
            type(value)
        ), "The `map` attribute must be a `starry` map instance."
        assert (
            value._lazy == self._lazy
        ), "Map must have the same evaluation mode (lazy/greedy)."
        self._map = value

    @property
    def r(self):
        """The radius in units of :py:attr:`length_unit`."""
        return self._r / self._length_factor

    @r.setter
    def r(self, value):
        self._r = self._math.cast(value * self._length_factor)

    @property
    def m(self):
        """The mass in units of :py:attr:`mass_unit`."""
        return self._m / self._mass_factor

    @m.setter
    def m(self, value):
        self._m = self._math.cast(value * self._mass_factor)

    @property
    def prot(self):
        """The rotation period in units of :py:attr:`time_unit`."""
        return self._prot / self._time_factor

    @prot.setter
    def prot(self, value):
        self._prot = self._math.cast(value * self._time_factor)

    @property
    def t0(self):
        """A reference time in units of :py:attr:`time_unit`."""
        return self._t0 / self._time_factor

    @t0.setter
    def t0(self, value):
        self._t0 = self._math.cast(value * self._time_factor)

    @property
    def theta0(self):
        """The map rotational phase at time :py:attr:`t0`."""
        return self._theta0 / self._angle_factor

    @theta0.setter
    def theta0(self, value):
        self._theta0 = self._math.cast(value * self._angle_factor)

    def _check_kwargs(self, method, kwargs):
        if not config.quiet:
            for key in kwargs.keys():
                message = "Invalid keyword `{0}` in call to `{1}()`. Ignoring."
                message = message.format(key, method)
                logger.warning(message)


class Primary(Body):
    """A primary (central) body.

    Args:
        map: The surface map of this body. This should be an instance
            returned by :py:func:`starry.Map`.
        r (scalar, optional): The radius of the body in units of
            :py:attr:`length_unit`. Defaults to 1.0.
        m (scalar, optional): The mass of the body in units of
            :py:attr:`mass_unit`. Defaults to 1.0.
        prot (scalar, optional): The rotation period of the body in units of
            :py:attr:`time_unit`. Defaults to 1.0.
        t0 (scalar, optional): A reference time in units of
            :py:attr:`time_unit`. Defaults to 0.0.
        theta0 (scalar, optional): The rotational phase of the map at time
            :py:attr:`t0` in units of :py:attr:`angle_unit`. Defaults to 0.0.
        length_unit (optional): An ``astropy.units`` unit defining the
            distance metric for this object. Defaults to
            :py:attr:`astropy.units.Rsun.`
        mass_unit (optional): An ``astropy.units`` unit defining the
            mass metric for this object. Defaults to
            :py:attr:`astropy.units.Msun.`
        time_unit (optional): An ``astropy.units`` unit defining the
            time metric for this object. Defaults to
            :py:attr:`astropy.units.day.`
        angle_unit (optional): An ``astropy.units`` unit defining the
            angular metric for this object. Defaults to
            :py:attr:`astropy.units.degree.`
    """

    def __init__(self, map, **kwargs):
        # Initialize `Body`
        super(Primary, self).__init__(map, **kwargs)
        for kw in [
            "r",
            "m",
            "prot",
            "t0",
            "theta0",
            "length_unit",
            "mass_unit",
            "time_unit",
            "angle_unit",
        ]:
            kwargs.pop(kw, None)
        self._check_kwargs("Primary", kwargs)


class Secondary(Body):
    """A secondary (orbiting) body.

    Args:
        map: The surface map of this body. This should be an instance
            returned by :py:func:`starry.Map`.
        r (scalar, optional): The radius of the body in units of
            :py:attr:`length_unit`. Defaults to 1.0.
        m (scalar, optional): The mass of the body in units of
            :py:attr:`mass_unit`. Defaults to 1.0.
        a (scalar, optional): The semi-major axis of the body in units of
            :py:attr:`time_unit`. Defaults to 1.0. If :py:attr:`porb` is
            also provided, this value is ignored.
        porb (scalar, optional): The orbital period of the body in units of
            :py:attr:`time_unit`. Defaults to 1.0. Setting this value
            overrides :py:attr:`a`.
        prot (scalar, optional): The rotation period of the body in units of
            :py:attr:`time_unit`. Defaults to 1.0.
        t0 (scalar, optional): A reference time in units of
            :py:attr:`time_unit`. This is taken to be the time of a reference
            transit. Defaults to 0.0.
        ecc (scalar, optional): The orbital eccentricity of the body.
            Defaults to 0.
        w, omega (scalar, optional): The argument of pericenter of the body
            in units of :py:attr:`angle_unit`. Defaults to 90 degrees.
        Omega (scalar, optional): The longitude of ascending node of the
            body in units of :py:attr:`angle_unit`. Defaults to 0 degrees.
        inc (scalar, optional): The orbital inclination of the body in
            units of :py:attr:`angle_unit`. Defaults to 90 degrees.
        theta0 (scalar, optional): The rotational phase of the map at time
            :py:attr:`t0` in units of :py:attr:`angle_unit`. Defaults to
            0.0.
        length_unit (optional): An ``astropy.units`` unit defining the
            distance metric for this object. Defaults to
            :py:attr:`astropy.units.Rsun.`
        mass_unit (optional): An ``astropy.units`` unit defining the
            mass metric for this object. Defaults to
            :py:attr:`astropy.units.Msun.`
        time_unit (optional): An ``astropy.units`` unit defining the
            time metric for this object. Defaults to
            :py:attr:`astropy.units.day.`
        angle_unit (optional): An ``astropy.units`` unit defining the
            angular metric for this object. Defaults to
            :py:attr:`astropy.units.degree.`
    """

    def __init__(self, map, **kwargs):
        # Initialize `Body`
        super(Secondary, self).__init__(map, **kwargs)
        for kw in [
            "r",
            "m",
            "prot",
            "t0",
            "theta0",
            "length_unit",
            "mass_unit",
            "time_unit",
            "angle_unit",
        ]:
            kwargs.pop(kw, None)

        # Attributes
        if kwargs.get("porb", None) is not None:
            self.porb = kwargs.pop("porb", None)
        elif kwargs.get("a", None) is not None:
            self.a = kwargs.pop("a", None)
        else:
            raise ValueError("Must provide a value for either `porb` or `a`.")
        self.ecc = kwargs.pop("ecc", 0.0)
        self.w = kwargs.pop(
            "w", kwargs.pop("omega", 0.5 * np.pi / self._angle_factor)
        )
        self.Omega = kwargs.pop("Omega", 0.0)
        self.inc = kwargs.pop("inc", 0.5 * np.pi / self._angle_factor)
        self._check_kwargs("Secondary", kwargs)

    @property
    def porb(self):
        """The orbital period in units of :py:attr:`time_unit`.

        .. note::
            Setting this value overrides the value of :py:attr:`a`.
        """
        if self._porb == 0.0:
            return None
        else:
            return self._porb / self._time_factor

    @porb.setter
    def porb(self, value):
        self._porb = self._math.cast(value * self._time_factor)
        self._a = 0.0

    @property
    def a(self):
        """The semi-major axis in units of :py:attr:`length_unit`.

        .. note::
            Setting this value overrides the value of :py:attr:`porb`.
        """
        if self._a == 0.0:
            return None
        else:
            return self._a / self._length_factor

    @a.setter
    def a(self, value):
        self._a = self._math.cast(value * self._length_factor)
        self._porb = 0.0

    @property
    def ecc(self):
        """The orbital eccentricity."""
        return self._ecc

    @ecc.setter
    def ecc(self, value):
        self._ecc = value

    @property
    def w(self):
        """The longitude of pericenter in units of :py:attr:`angle_unit`."""
        return self._w / self._angle_factor

    @w.setter
    def w(self, value):
        self._w = self._math.cast(value * self._angle_factor)

    @property
    def omega(self):
        """Alias for the longitude of pericenter :py:attr:`w`."""
        return self.w

    @omega.setter
    def omega(self, value):
        self.w = value

    @property
    def Omega(self):
        """The longitude of ascending node in units of :py:attr:`angle_unit`."""
        return self._Omega / self._angle_factor

    @Omega.setter
    def Omega(self, value):
        self._Omega = self._math.cast(value * self._angle_factor)

    @property
    def inc(self):
        """The orbital inclination in units of :py:attr:`angle_unit`."""
        return self._inc / self._angle_factor

    @inc.setter
    def inc(self, value):
        self._inc = self._math.cast(value * self._angle_factor)


class System(object):
    """A system of bodies in Keplerian orbits about a central primary body.

    Args:
        primary (:py:class:`Primary`): The central body.
        secondaries (:py:class:`Secondary`): One or more secondary bodies
            in orbit about the primary.
        time_unit (optional): An ``astropy.units`` unit defining the
            time metric for this object. Defaults to
            :py:attr:`astropy.units.day.`
        light_delay (bool, optional): Account for the light travel time
            delay to the barycenter of the system? Default is False.
        texp (scalar): The exposure time of each observation. This can be a
            scalar or a tensor with the same shape as ``t``. If ``texp`` is
            provided, ``t`` is assumed to indicate the timestamp at the middle
            of an exposure of length ``texp``.
        oversample (int): The number of function evaluations to use when
            numerically integrating the exposure time.
        order (int): The order of the numerical integration scheme. This must
            be one of the following: ``0`` for a centered Riemann sum
            (equivalent to the "resampling" procedure suggested by Kipping 2010),
            ``1`` for the trapezoid rule, or ``2`` for Simpson’s rule.
    """

    def _no_spectral(self):
        if self._primary._map.nw is not None:  # pragma: no cover
            raise NotImplementedError(
                "Method not yet implemented for spectral maps."
            )

    def __init__(
        self,
        primary,
        *secondaries,
        time_unit=units.day,
        light_delay=False,
        texp=None,
        oversample=7,
        order=0,
    ):
        # Units
        self.time_unit = time_unit
        self._light_delay = bool(light_delay)
        if texp is None:
            self._texp = 0.0
        else:
            self._texp = texp
        assert self._texp >= 0.0, "Parameter `texp` must be >= 0."
        self._oversample = int(oversample)
        assert self._oversample > 0, "Parameter `oversample` must be > 0."
        self._order = int(order)
        assert self._order in [0, 1, 2], "Invalid value for parameter `order`."

        # Primary body
        assert (
            type(primary) is Primary
        ), "Argument `primary` must be an instance of `Primary`."
        assert (
            primary._map.__props__["reflected"] == False
        ), "Reflected light map not allowed for the primary body."
        self._primary = primary
        self._rv = primary._map.__props__["rv"]
        self._lazy = primary._lazy
        self._math = primary._math
        if self._lazy:
            self._linalg = math.lazy_linalg
        else:
            self._linalg = math.greedy_linalg

        # Secondary bodies
        assert len(secondaries) > 0, "There must be at least one secondary."
        for sec in secondaries:
            assert type(sec) is Secondary, (
                "Argument `*secondaries` must be a sequence of "
                "`Secondary` instances."
            )
            assert (
                sec._map.nw == self._primary._map.nw
            ), "All bodies must have the same number of wavelength bins `nw`."
            assert sec._map.__props__["rv"] == self._rv, (
                "Radial velocity must be enabled "
                "for either all or none of the bodies."
            )
            assert (
                sec._lazy == self._lazy
            ), "All bodies must have the same evaluation mode (lazy/greedy)."

        reflected = [sec._map.__props__["reflected"] for sec in secondaries]
        if np.all(reflected):
            self._reflected = True
        elif np.any(reflected):
            raise ValueError(
                "Reflected light must be enabled "
                "for either all or none of the secondaries."
            )
        else:
            self._reflected = False
        self._secondaries = secondaries

        # All bodies
        self._bodies = [self._primary] + list(self._secondaries)

        # Indices of each of the bodies in the design matrix
        Ny = [self._primary._map.Ny] + [
            sec._map.Ny for sec in self._secondaries
        ]
        self._inds = []
        cur = 0
        for N in Ny:
            self._inds.append(cur + np.arange(N))
            cur += N

        # Theano ops class
        self.ops = OpsSystem(
            self._primary,
            self._secondaries,
            reflected=self._reflected,
            rv=self._rv,
            light_delay=self._light_delay,
            texp=self._texp,
            oversample=self._oversample,
            order=self._order,
        )

        # Solve stuff
        self._flux = None
        self._C = None
        self._solution = None
        self._solved_bodies = []

    @property
    def light_delay(self):
        """Account for the light travel time delay? *Read-only*"""
        return self._light_delay

    @property
    def texp(self):
        """The exposure time in units of :py:attr:`time_unit`. *Read-only*"""

    @property
    def oversample(self):
        """Oversample factor when integrating over exposure time. *Read-only*"""
        return self._oversample

    @property
    def order(self):
        """The order of the numerical integration scheme. *Read-only*

        - ``0``: a centered Riemann sum
        - ``1``: trapezoid rule
        - ``2``: Simpson’s rule
        """
        return self._order

    @property
    def time_unit(self):
        """An ``astropy.units`` unit defining the time metric for the system."""
        return self._time_unit

    @time_unit.setter
    def time_unit(self, value):
        assert value.physical_type == "time"
        self._time_unit = value
        self._time_factor = value.in_units(units.day)

    @property
    def primary(self):
        """The primary (central) object in the Keplerian system."""
        return self._primary

    @property
    def secondaries(self):
        """A list of the secondary (orbiting) object(s) in the Keplerian system."""
        return self._secondaries

    @property
    def bodies(self):
        """A list of all objects in the Keplerian system."""
        return self._bodies

    @property
    def map_indices(self):
        """A list of the indices corresponding to each body in the design matrix."""
        return self._inds

    def show(
        self,
        t,
        cmap="plasma",
        res=300,
        interval=75,
        file=None,
        figsize=(3, 3),
        html5_video=True,
        window_pad=1.0,
        **kwargs,
    ):
        """Visualize the Keplerian system.

        Note that the body surface intensities are not normalized.

        Args:
            t (scalar or vector): The time(s) at which to evaluate the orbit and
                the map in units of :py:attr:`time_unit`.
            cmap (string or colormap instance, optional): The matplotlib colormap
                to use. Defaults to ``plasma``.
            res (int, optional): The resolution of the map in pixels on a
                side. Defaults to 300.
            figsize (tuple, optional): Figure size in inches. Default is
                (3, 3) for orthographic maps and (7, 3.5) for rectangular
                maps.
            interval (int, optional): Interval between frames in milliseconds
                (animated maps only). Defaults to 75.
            file (string, optional): The file name (including the extension)
                to save the animation to (animated maps only). Defaults to None.
            html5_video (bool, optional): If rendering in a Jupyter notebook,
                display as an HTML5 video? Default is True. If False, displays
                the animation using Javascript (file size will be larger.)
            window_pad (float, optional): Padding around the primary in units
                of the primary radius. Bodies outside of this window will be
                cropped. Default is 1.0.

        .. note::
            If calling this method on an instance of ``System`` created within
            a ``pymc3.Model()``, you may specify a ``point`` keyword with
            the model point at which to evaluate the map. This method also
            accepts a ``model`` keyword, although this is inferred
            automatically if called from within a ``pymc3.Model()`` context.
            If no point is provided, attempts to evaluate the map at
            ``model.test_point`` and raises a warning.

        """
        # Not yet implemented
        if self._primary._map.nw is not None:  # pragma: no cover
            raise NotImplementedError(
                "Method not implemented for spectral maps."
            )

        # So we can evaluate stuff in lazy mode
        get_val = evaluator(**kwargs)

        # Render the maps & get the orbital positions
        if self._rv:
            self._primary.map._set_RV_filter()
            for sec in self._secondaries:
                sec.map._set_RV_filter()
        img_pri, img_sec, x, y, z = self.ops.render(
            self._math.reshape(self._math.to_array_or_tensor(t), [-1])
            * self._time_factor,
            res,
            self._primary._r,
            self._primary._m,
            self._primary._prot,
            self._primary._t0,
            self._primary._theta0,
            self._primary._map._inc,
            self._primary._map._obl,
            self._primary._map._y,
            self._primary._map._u,
            self._primary._map._f,
            self._math.to_array_or_tensor(
                [sec._r for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._m for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._prot for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._t0 for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._theta0 for sec in self._secondaries]
            ),
            self._get_periods(),
            self._math.to_array_or_tensor(
                [sec._ecc for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._w for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._Omega for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._inc for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._map._inc for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._map._obl for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._map._y for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._map._u for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._map._f for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._map._sigr for sec in self._secondaries]
            ),
        )

        # Convert to units of the primary radiu
        x, y, z = (
            x / self._primary._r,
            y / self._primary._r,
            z / self._primary._r,
        )
        r = self._math.to_array_or_tensor(
            [sec._r for sec in self._secondaries]
        )
        r = r / self._primary._r

        # Evaluate if needed
        if self._lazy:
            img_pri = get_val(img_pri)
            img_sec = get_val(img_sec)
            x = get_val(x)
            y = get_val(y)
            z = get_val(z)
            r = get_val(r)

        # We need this to be of shape (nplanet, nframe)
        x = x.T
        y = y.T
        z = z.T

        # Ensure we have an array of frames
        if len(img_pri.shape) == 3:
            nframes = img_pri.shape[0]
        else:  # pragma: no cover
            nframes = 1
            img_pri = np.reshape(img_pri, (1,) + img_pri.shape)
            img_sec = np.reshape(img_sec, (1,) + img_sec.shape)
        animated = nframes > 1

        # Set up the plot
        fig, ax = plt.subplots(1, figsize=figsize)
        ax.axis("off")
        ax.set_xlim(-1.0 - window_pad, 1.0 + window_pad)
        ax.set_ylim(-1.0 - window_pad, 1.0 + window_pad)

        # Render the first frame
        img = [None for n in range(1 + len(self._secondaries))]
        circ = [None for n in range(1 + len(self._secondaries))]
        extent = np.array([-1.0, 1.0, -1.0, 1.0])
        img[0] = ax.imshow(
            img_pri[0],
            origin="lower",
            extent=extent,
            cmap=cmap,
            interpolation="none",
            vmin=np.nanmin(img_pri),
            vmax=np.nanmax(img_pri),
            animated=animated,
            zorder=0.0,
        )
        circ[0] = plt.Circle(
            (0, 0), 1, color="k", fill=False, zorder=1e-3, lw=2
        )
        ax.add_artist(circ[0])
        for i, _ in enumerate(self._secondaries):
            extent = np.array([x[i, 0], x[i, 0], y[i, 0], y[i, 0]]) + (
                r[i] * np.array([-1.0, 1.0, -1.0, 1.0])
            )
            img[i + 1] = ax.imshow(
                img_sec[i, 0],
                origin="lower",
                extent=extent,
                cmap=cmap,
                interpolation="none",
                vmin=np.nanmin(img_sec),
                vmax=np.nanmax(img_sec),
                animated=animated,
                zorder=z[i, 0],
            )
            circ[i] = plt.Circle(
                (x[i, 0], y[i, 0]),
                r[i],
                color="k",
                fill=False,
                zorder=z[i, 0] + 1e-3,
                lw=2,
            )
            ax.add_artist(circ[i])

        # Animation
        if animated:

            def updatefig(k):

                # Update Primary map
                img[0].set_array(img_pri[k])

                # Update Secondary maps & positions
                for i, _ in enumerate(self._secondaries):
                    extent = np.array([x[i, k], x[i, k], y[i, k], y[i, k]]) + (
                        r[i] * np.array([-1.0, 1.0, -1.0, 1.0])
                    )
                    if np.any(np.abs(extent) < 1.0 + window_pad):
                        img[i + 1].set_array(img_sec[i, k])
                        img[i + 1].set_extent(extent)
                        img[i + 1].set_zorder(z[i, k])
                        circ[i].center = (x[i, k], y[i, k])
                        circ[i].set_zorder(z[i, k] + 1e-3)

                return img + circ

            ani = FuncAnimation(
                fig, updatefig, interval=interval, blit=False, frames=nframes
            )

            # Business as usual
            if (file is not None) and (file != ""):
                if file.endswith(".mp4"):
                    ani.save(file, writer="ffmpeg")
                elif file.endswith(".gif"):
                    ani.save(file, writer="imagemagick")
                else:  # pragma: no cover
                    # Try and see what happens!
                    ani.save(file)
                plt.close()
            else:  # pragma: no cover
                try:
                    if "zmqshell" in str(type(get_ipython())):
                        plt.close()
                        if html5_video:
                            display(HTML(ani.to_html5_video()))
                        else:
                            display(HTML(ani.to_jshtml()))
                    else:
                        raise NameError("")
                except NameError:
                    plt.show()
                    plt.close()

            # Matplotlib generates an annoying empty
            # file when producing an animation. Delete it.
            try:
                os.remove("None0000000.png")
            except FileNotFoundError:
                pass

        else:

            if (file is not None) and (file != ""):
                fig.savefig(file)
                plt.close()
            else:  # pragma: no cover
                plt.show()

        if self._rv:
            self._primary.map._unset_RV_filter()
            for sec in self._secondaries:
                sec.map._unset_RV_filter()

    def design_matrix(self, t):
        """Compute the system flux design matrix at times ``t``.

        .. note::

            This is the *unweighted* design matrix, i.e., it does not
            include the scaling by the amplitude of each body's map.
            To perform this weighting, do

            .. code-block:: python

                X = sys.design_matrix(**kwargs)
                for i, body in zip(sys.map_indices, sys.bodies):
                    X[:, i] *= body.map.amp

        Args:
            t (scalar or vector): An array of times at which to evaluate
                the design matrix in units of :py:attr:`time_unit`.
        """
        return self.ops.X(
            self._math.reshape(self._math.to_array_or_tensor(t), [-1])
            * self._time_factor,
            self._primary._r,
            self._primary._m,
            self._primary._prot,
            self._primary._t0,
            self._primary._theta0,
            self._math.to_array_or_tensor(1.0),
            self._primary._map._inc,
            self._primary._map._obl,
            self._primary._map._u,
            self._primary._map._f,
            self._math.to_array_or_tensor(
                [sec._r for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._m for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._prot for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._t0 for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._theta0 for sec in self._secondaries]
            ),
            self._get_periods(),
            self._math.to_array_or_tensor(
                [sec._ecc for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._w for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._Omega for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._inc for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [
                    self._math.to_array_or_tensor(1.0)
                    for sec in self._secondaries
                ]
            ),
            self._math.to_array_or_tensor(
                [sec._map._inc for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._map._obl for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._map._u for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._map._f for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._map._sigr for sec in self._secondaries]
            ),
        )

    def flux(self, t, total=True):
        """Compute the system flux at times ``t``.

        Args:
            t (scalar or vector): An array of times at which to evaluate
                the flux in units of :py:attr:`time_unit`.
            total (bool, optional): Return the total system flux? Defaults to
                True. If False, returns arrays corresponding to the flux
                from each body.
        """
        X = self.design_matrix(t)

        # Weight the ylms by amplitude
        if self._reflected:
            # If we're doing reflected light, scale the amplitude of
            # each of the secondaries by the amplitude of the primary
            # (the illumination source).
            ay = [self._primary.map.amp * self._primary._map._y] + [
                self._primary.map.amp * body.map.amp * body._map._y
                for body in self._secondaries
            ]
        else:
            ay = [body.map.amp * body._map._y for body in self._bodies]

        if total:
            return self._math.dot(X, self._math.concatenate(ay))
        else:
            return [
                self._math.dot(X[:, idx], ay[i])
                for i, idx in enumerate(self._inds)
            ]

    def rv(self, t, keplerian=True, total=True):
        """Compute the observed radial velocity of the system at times ``t``.

        Args:
            t (scalar or vector): An array of times at which to evaluate
                the radial velocity in units of :py:attr:`time_unit`.
            keplerian (bool): Include the Keplerian component of the radial
                velocity of the primary? Default is True. If False, this
                method returns a model for only the radial velocity anomaly
                due to transits (the Rossiter-McLaughlin effect) and
                time-variable surface features (Doppler tomography) for all
                bodies in the system.
            total (bool, optional): Return the total system RV? Defaults to
                True. If False, returns arrays corresponding to the RV
                contribution from each body.

        Returns:
            A vector or matrix of radial velocities in units of meters per
            second. If ``total`` is ``True``, this returns a vector of the
            total stellar radial velocity at every point ``t``. This is
            the sum of all effects: the Keplerian motion of the star due to
            the planets, the radial velocity anomaly due to spots or
            features on its surface rotating in and out of view, as well
            as the Rossiter-McLaughlin effect due to transits by the
            secondaries.
            If ``total`` is ``False``, this returns a matrix whose
            rows are the radial velocity contributions to the measured **stellar**
            radial velocity from each of the bodies in the system.
            As a specific example, if there are three bodies in the
            system (one ``Primary`` and two ``Secondary`` bodies), the
            first row is the radial velocity anomaly corresponding to the star's
            own Doppler signals (due to spots rotating in and out of
            view, or due to the Rossiter-McLaughlin effect of the other
            two bodies transiting it). The other two rows are the Keplerian
            contribution from each of the ``Secondary`` bodies, plus any
            radial velocity anomaly due to spots or occultations of their
            surfaces.
            The sum across all rows is equal to the total radial velocity
            and is the same as what this method would return if ``total=True``.

        Note, importantly, that when ``total=False``, this method does
        *not* return the radial velocities of each of the bodies; instead,
        it returns the *contribution to the stellar radial velocity* due
        to each of the bodies. If you require knowing the radial velocity
        of the secondary objects, you can compute this from conservation
        of momentum:

        .. code-block::python

            # Instantiate a system w/ a star and two planets
            A = starry.Primary(...)
            b = starry.Secondary(...)
            c = starry.Secondary(...)
            sys = starry.System(A, b, c)

            # Get the contribution from each body to the star's RV
            rv = sys.rv(t, total=False)

            # Conservation of momentum implies the RV of `b`
            # is proportional to the RV of the star, weighted
            # by the mass ratio. We can compute the mass ratio
            # if we're mindful of the (potentially different)
            # units for the star and the planets:
            mA_mb = ((A.m * A.mass_unit) / (b.m * b.mass_unit)).decompose()
            rv_b = -rv[1] * mA_mb

            # Same for planet `c`
            mA_mc = ((A.m * A.mass_unit) / (b.m * b.mass_unit)).decompose()
            rv_c = -rv[2] * mAmc

        Note that this method implicitly assumes multi-Keplerian orbits;
        i.e., the ``Secondary`` bodies are treated as massive *only* when computing
        their gravitational effect on the ``Primary`` (as opposed to each other).
        This therefore ignores all ``Secondary``-``Secondary``
        (i.e., planet-planet) interactions.

        """
        assert self._rv, "Only implemented if `rv=True` for all body maps."
        rv = self.ops.rv(
            self._math.reshape(self._math.to_array_or_tensor(t), [-1])
            * self._time_factor,
            self._primary._r,
            self._primary._m,
            self._primary._prot,
            self._primary._t0,
            self._primary._theta0,
            self._primary._map._amp,
            self._primary._map._inc,
            self._primary._map._obl,
            self._primary._map._y,
            self._primary._map._u,
            self._primary._map._alpha,
            self._primary._map._veq,
            self._math.to_array_or_tensor(
                [sec._r for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._m for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._prot for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._t0 for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._theta0 for sec in self._secondaries]
            ),
            self._get_periods(),
            self._math.to_array_or_tensor(
                [sec._ecc for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._w for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._Omega for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._inc for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._map._amp for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._map._inc for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._map._obl for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._map._y for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._map._u for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._map._alpha for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._map._sigr for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._map._veq for sec in self._secondaries]
            ),
            np.array(keplerian),
        )
        if total:
            return self._math.sum(rv, axis=0)
        else:
            return rv

    def position(self, t):
        """Compute the Cartesian positions of all bodies at times ``t``.

        Args:
            t (scalar or vector): An array of times at which to evaluate
                the position in units of :py:attr:`time_unit`.
        """
        x, y, z = self.ops.position(
            self._math.reshape(self._math.to_array_or_tensor(t), [-1])
            * self._time_factor,
            self._primary._m,
            self._primary._t0,
            self._math.to_array_or_tensor(
                [sec._m for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._t0 for sec in self._secondaries]
            ),
            self._get_periods(),
            self._math.to_array_or_tensor(
                [sec._ecc for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._w for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._Omega for sec in self._secondaries]
            ),
            self._math.to_array_or_tensor(
                [sec._inc for sec in self._secondaries]
            ),
        )
        fac = np.reshape(
            [self._primary._length_factor]
            + [sec._length_factor for sec in self._secondaries],
            [-1, 1],
        )
        return (x / fac, y / fac, z / fac)

    def _get_periods(self):
        periods = [None for sec in self._secondaries]
        for i, sec in enumerate(self._secondaries):
            if sec._porb:
                periods[i] = sec._porb
            else:
                periods[i] = (
                    (2 * np.pi)
                    * sec._a ** (3 / 2)
                    / (self._math.sqrt(G_grav * (self._primary._m + sec._m)))
                )
        return self._math.to_array_or_tensor(periods)

    def set_data(self, flux, C=None, cho_C=None):
        """Set the data vector and covariance matrix.

        This method is required by the :py:meth:`solve` method, which
        analytically computes the posterior over surface maps for all bodies
        in the system given a dataset and a prior, provided both are described
        as multivariate Gaussians.

        Args:
            flux (vector): The observed system light curve.
            C (scalar, vector, or matrix): The data covariance. This may be
                a scalar, in which case the noise is assumed to be
                homoscedastic, a vector, in which case the covariance
                is assumed to be diagonal, or a matrix specifying the full
                covariance of the dataset. Default is None. Either `C` or
                `cho_C` must be provided.
            cho_C (matrix): The lower Cholesky factorization of the data
                covariance matrix. Defaults to None. Either `C` or
                `cho_C` must be provided.
        """
        self._flux = self._math.cast(flux)
        self._C = self._linalg.Covariance(
            C=C, cho_C=cho_C, N=self._flux.shape[0]
        )

    def solve(self, *, design_matrix=None, t=None):
        """Solve the least-squares problem for the posterior over maps for all bodies.

        This method solves the generalized least squares problem given a system
        light curve and its covariance (set via the :py:meth:`set_data` method)
        and a Gaussian prior on the spherical harmonic coefficients
        (set via the :py:meth:`set_prior` method). The map amplitudes and
        coefficients of each of the bodies in the system are then set to the
        maximum a posteriori (MAP) solution.

        Args:
            design_matrix (matrix, optional): The flux design matrix, the
                quantity returned by :py:meth:`design_matrix`. Default is
                None, in which case this is computed based on ``kwargs``.
            t (vector, optional): The vector of times at which to evaluate
                :py:meth:`design_matrix`, if a design matrix is not provided.
                Default is None.

        Returns:
            The posterior mean for the spherical harmonic \
            coefficients `l > 0` and the Cholesky factorization of the \
            posterior covariance of all of the bodies in the system, \
            stacked in order (primary, followed by each of the secondaries \
            in the order they were provided.)

        .. note::
            Users may call the :py:meth:`draw` method of this class to draw
            from the posterior after calling :py:meth:`solve`.
        """
        # TODO: Implement for spectral maps?
        self._no_spectral()

        # Check that the data is set
        if self._flux is None or self._C is None:
            raise ValueError("Please provide a dataset with `set_data()`.")

        # Get the full design matrix
        if design_matrix is None:
            assert t is not None, "Please provide a time vector `t`."
            design_matrix = self.design_matrix(t)
        X = self._math.cast(design_matrix)

        # Get the data vector
        f = self._math.cast(self._flux)

        # Check for bodies whose priors are set
        self._solved_bodies = []
        inds = []
        dense_L = False
        for k, body in enumerate(self._bodies):

            if body.map._mu is None or body.map._L is None:

                # Subtract out this term from the data vector,
                # since it is fixed
                f -= body.map.amp * self._math.dot(
                    X[:, self._inds[k]], body.map.y
                )

            else:

                # Add to our list of indices/bodies to solve for
                inds.extend(self._inds[k])
                self._solved_bodies.append(body)
                if body.map._L.kind in ["matrix", "cholesky"]:
                    dense_L = True

        # Do we have at least one body?
        if len(self._solved_bodies) == 0:
            raise ValueError("Please provide a prior for at least one body.")

        # Keep only the terms we'll solve for
        X = X[:, inds]

        # Stack our priors
        mu = self._math.concatenate(
            [body.map._mu for body in self._solved_bodies]
        )

        if not dense_L:
            # We can just concatenate vectors
            LInv = self._math.concatenate(
                [
                    body.map._L.inverse * self._math.ones(body.map.Ny)
                    for body in self._solved_bodies
                ]
            )
        else:
            # FACT: The inverse of a block diagonal matrix
            # is the block diagonal matrix of the inverses.
            LInv = self._math.block_diag(
                *[
                    body.map._L.inverse * self._math.eye(body.map.Ny)
                    for body in self._solved_bodies
                ]
            )

        # Compute the MAP solution
        self._solution = self._linalg.solve(X, f, self._C.cholesky, mu, LInv)

        # Set all the map vectors
        x, cho_cov = self._solution
        n = 0
        for body in self._solved_bodies:
            inds = slice(n, n + body.map.Ny)
            body.map.amp = x[inds][0]
            if body.map.ydeg > 0:
                body.map[1:, :] = x[inds][1:] / body.map.amp
            n += body.map.Ny

        # Return the mean and covariance
        self._solution = (x, cho_cov)
        return self._solution

    @property
    def solution(self):
        r"""The posterior probability distribution for the maps in the system.

        This is a tuple containing the mean and lower Cholesky factorization of the
        covariance of the amplitude-weighted spherical harmonic coefficient vectors,
        obtained by solving the regularized least-squares problem
        via the :py:meth:`solve` method.

        Note that to obtain the actual covariance matrix from the lower Cholesky
        factorization :math:`L`, simply compute :math:`L L^\top`.

        Note also that this is the posterior for the **amplitude-weighted**
        map vectors. Under this convention, the map amplitude is equal to the
        first term of the vector of each body and the spherical harmonic coefficients are
        equal to the vector normalized by the first term.
        """
        if self._solution is None:
            raise ValueError("Please call `solve()` first.")
        return self._solution

    def draw(self):
        """
        Draw a map from the posterior distribution and set
        the :py:attr:`y` map vector of each body.

        Users should call :py:meth:`solve` to enable this attribute.
        """
        if self._solution is None:
            raise ValueError("Please call `solve()` first.")

        # Number of coefficients
        N = np.sum([body.map.Ny for body in self._solved_bodies])

        # Fast multivariate sampling using the Cholesky factorization
        yhat, cho_ycov = self._solution
        u = self._math.cast(np.random.randn(N))
        x = yhat + self._math.dot(cho_ycov, u)

        # Set all the map vectors
        n = 0
        for body in self._solved_bodies:
            inds = slice(n, n + body.map.Ny)
            body.map.amp = x[inds][0]
            body.map[1:, :] = x[inds][1:] / body.map.amp
            n += body.map.Ny

    def lnlike(self, *, design_matrix=None, t=None, woodbury=True):
        """Returns the log marginal likelihood of the data given a design matrix.

        This method computes the marginal likelihood (marginalized over the
        spherical harmonic coefficients of all bodies) given a system
        light curve and its covariance (set via the :py:meth:`set_data` method)
        and a Gaussian prior on the spherical harmonic coefficients
        (set via the :py:meth:`set_prior` method).

        Args:
            design_matrix (matrix, optional): The flux design matrix, the
                quantity returned by :py:meth:`design_matrix`. Default is
                None, in which case this is computed based on ``kwargs``.
            t (vector, optional): The vector of times at which to evaluate
                :py:meth:`design_matrix`, if a design matrix is not provided.
                Default is None.
            woodbury (bool, optional): Solve the linear problem using the
                Woodbury identity? Default is True. The
                `Woodbury identity <https://en.wikipedia.org/wiki/Woodbury_matrix_identity>`_
                is used to speed up matrix operations in the case that the
                number of data points is much larger than the number of
                spherical harmonic coefficients. In this limit, it can
                speed up the code by more than an order of magnitude. Keep
                in mind that the numerical stability of the Woodbury identity
                is not great, so if you're getting strange results try
                disabling this. It's also a good idea to disable this in the
                limit of few data points and large spherical harmonic degree.

        Returns:
            lnlike: The log marginal likelihood.
        """
        # TODO: Implement for spectral maps?
        self._no_spectral()

        # Check that the data is set
        if self._flux is None or self._C is None:
            raise ValueError("Please provide a dataset with `set_data()`.")

        # Get the full design matrix
        if design_matrix is None:
            assert t is not None, "Please provide a time vector `t`."
            design_matrix = self.design_matrix(t)
        X = self._math.cast(design_matrix)

        # Get the data vector
        f = self._math.cast(self._flux)

        # Check for bodies whose priors are set
        self._solved_bodies = []
        inds = []
        dense_L = False
        for k, body in enumerate(self._bodies):

            if body.map._mu is None or body.map._L is None:

                # Subtract out this term from the data vector,
                # since it is fixed
                f -= body.map.amp * self._math.dot(
                    X[:, self._inds[k]], body.map.y
                )

            else:

                # Add to our list of indices/bodies to solve for
                inds.extend(self._inds[k])
                self._solved_bodies.append(body)
                if body.map._L.kind in ["matrix", "cholesky"]:
                    dense_L = True

        # Do we have at least one body?
        if len(self._solved_bodies) == 0:
            raise ValueError("Please provide a prior for at least one body.")

        # Keep only the terms we'll solve for
        X = X[:, inds]

        # Stack our priors
        mu = self._math.concatenate(
            [body.map._mu for body in self._solved_bodies]
        )

        # Compute the likelihood
        if woodbury:
            if not dense_L:
                # We can just concatenate vectors
                LInv = self._math.concatenate(
                    [
                        body.map._L.inverse * self._math.ones(body.map.Ny)
                        for body in self._solved_bodies
                    ]
                )
            else:
                LInv = self._math.block_diag(
                    *[
                        body.map._L.inverse * self._math.eye(body.map.Ny)
                        for body in self._solved_bodies
                    ]
                )
            lndetL = self._math.cast(
                [body.map._L.lndet for body in self._solved_bodies]
            )
            return self._linalg.lnlike_woodbury(
                X, f, self._C.inverse, mu, LInv, self._C.lndet, lndetL
            )
        else:
            if not dense_L:
                # We can just concatenate vectors
                L = self._math.concatenate(
                    [
                        body.map._L.value * self._math.ones(body.map.Ny)
                        for body in self._solved_bodies
                    ]
                )
            else:
                L = self._math.block_diag(
                    *[
                        body.map._L.value * self._math.eye(body.map.Ny)
                        for body in self._solved_bodies
                    ]
                )
            return self._linalg.lnlike(X, f, self._C.value, mu, L)



import starry
import matplotlib.pyplot as plt
import numpy as np
import pytest


# BROKEN. We need to re-parametrize
# the L and K integrals (or compute them numerically).
@pytest.mark.xfail
def test_high_l_stability(plot=False):
    map = starry.Map(ydeg=20, reflected=False)
    map[1:, :] = 1
    xo = np.logspace(-2, np.log10(2.0), 1000)
    yo = 0
    ro = 0.9
    flux = map.flux(xo=xo, yo=yo, ro=ro)
    bo = np.sqrt(xo ** 2 + yo ** 2)
    ksq = (1 - ro ** 2 - bo ** 2 + 2 * bo * ro) / (4 * bo * ro)

    if plot:
        plt.plot(ksq, flux)
        plt.xscale("log")
        plt.show()

    # Check for gross stability issues here
    assert np.std(flux[ksq > 1]) < 0.1


if __name__ == "__main__":
    starry.config.lazy = False
    plt.style.use("default")
    test_high_l_stability(plot=True)



# -*- coding: utf-8 -*-
"""
Miscellaneous tests.

"""
import starry
import io
import logging
import warnings


def test_quiet():
    """Test quiet setting."""
    # Get root logger
    logger = logging.getLogger()

    # Assert we get a message
    starry.config.quiet = False
    with io.StringIO() as capture:
        ch = logging.StreamHandler(capture)
        ch.setLevel(logging.DEBUG)
        logger.addHandler(ch)
        map = starry.Map(ydeg=1)
        assert "Pre-computing some matrices" in capture.getvalue()
        logger.removeHandler(ch)

    # Assert we don't
    starry.config.quiet = True
    with io.StringIO() as capture:
        ch = logging.StreamHandler(capture)
        ch.setLevel(logging.DEBUG)
        logger.addHandler(ch)
        map = starry.Map(ydeg=1)
        assert len(capture.getvalue()) == 0
        logger.removeHandler(ch)


def test_check_kwargs_map(caplog):
    """Test that we capture bad keyword arguments."""
    starry.config.quiet = False
    caplog.clear()
    map = starry.Map(giraffe=10)
    assert len(caplog.records) >= 1
    assert any(
        [
            "Invalid keyword `giraffe`" in str(rec.message)
            for rec in caplog.records
        ]
    )


def test_check_kwargs_map_flux(caplog):
    """Test that we capture bad keyword arguments."""
    starry.config.quiet = False
    caplog.clear()
    flux = starry.Map().flux(giraffe=10)
    assert len(caplog.records) >= 1
    assert any(
        [
            "Invalid keyword `giraffe`" in str(rec.message)
            for rec in caplog.records
        ]
    )


def test_check_kwargs_body(caplog):
    """Test that we capture bad keyword arguments."""
    starry.config.quiet = False
    caplog.clear()
    body = starry.Primary(starry.Map(), giraffe=10)
    assert len(caplog.records) >= 1
    assert any(
        [
            "Invalid keyword `giraffe`" in str(rec.message)
            for rec in caplog.records
        ]
    )


def test_coefficient_numbers():
    map = starry.Map(ydeg=2, udeg=3)
    assert map.ydeg == 2
    assert map.udeg == 3
    assert map.fdeg == 0
    assert map.deg == 5
    assert map.Ny == 9
    assert map.Nu == 4
    assert map.Nf == 1
    assert map.N == 36



# -*- coding: utf-8 -*-
"""Test light travel time delay"""
import starry
import numpy as np


def test_light_delay():
    pri = starry.Primary(starry.Map())
    sec = starry.Secondary(starry.Map(), porb=1.0)
    sys = starry.System(pri, sec, light_delay=True)
    assert sys.light_delay is True

    # TODO: Add tests here.



# -*- coding: utf-8 -*-
"""
Test various exceptions throughout the code.

"""
import starry
import pytest


def test_negative_radius():
    map = starry.Map(ydeg=2)
    map[1, 0] = 1.0
    with pytest.raises(RuntimeError) as e:
        flux = map.flux(xo=0.99, yo=0, zo=1, ro=-1e-15)
    assert "Occultor radius is negative. Aborting." in str(e.value)


def test_bad_ylm_indices():
    map = starry.Map(ydeg=1, udeg=1)

    # Bad `l`
    with pytest.raises(ValueError) as e:
        x = map[2, 0]
    with pytest.raises(ValueError) as e:
        map[2, 0] = 1

    # Bad `l`
    with pytest.raises(ValueError) as e:
        x = map[-1, 0]
    with pytest.raises(ValueError) as e:
        map[-1, 0] = 1

    # Bad `m`
    x = map[1, 3]
    assert len(x) == 0
    # TODO: It would be nice if `map[1, 3] = ...` raised
    # an error, but currently it does nothing (silently).

    # Bad type
    with pytest.raises(ValueError) as e:
        x = map[2.3, 5.7]
    with pytest.raises(ValueError) as e:
        map[2.3, 5.7] = 1


def test_bad_wavelength_indices():
    map = starry.Map(ydeg=1, udeg=1, nw=2)

    # Bad index
    with pytest.raises(ValueError) as e:
        x = map[1, 0, -1]
    with pytest.raises(ValueError) as e:
        map[1, 0, -1] = 1

    # Bad type
    with pytest.raises(ValueError) as e:
        x = map[1, 0, "foo"]
    with pytest.raises(ValueError) as e:
        map[1, 0, "foo"] = 1


def test_bad_ul_indices():
    map = starry.Map(ydeg=1, udeg=1)

    # Bad `l`
    with pytest.raises(ValueError) as e:
        x = map[-1]
    with pytest.raises(ValueError) as e:
        map[-1] = 1

    # Bad `l`
    with pytest.raises(ValueError) as e:
        x = map[3]
    with pytest.raises(ValueError) as e:
        map[3] = 1

    # Bad type
    with pytest.raises(ValueError) as e:
        x = map[map]
    with pytest.raises(ValueError) as e:
        map[map] = 1


def test_bad_sys_settings():
    pri = starry.Primary(starry.Map())
    sec = starry.Secondary(starry.Map(), porb=1.0)

    # Bad exposure time
    with pytest.raises(AssertionError) as e:
        sys = starry.System(pri, sec, texp=-1.0)

    # Bad oversample factor
    with pytest.raises(AssertionError) as e:
        sys = starry.System(pri, sec, oversample=-1)

    # Bad integration order
    with pytest.raises(AssertionError) as e:
        sys = starry.System(pri, sec, order=99)

    # Bad primary
    with pytest.raises(AssertionError) as e:
        sys = starry.System(sec, sec)

    # Reflected light primary
    with pytest.raises(AssertionError) as e:
        sys = starry.System(starry.Primary(starry.Map(reflected=True)), sec)

    # Bad secondary
    with pytest.raises(AssertionError) as e:
        sys = starry.System(pri, pri)

    # No secondaries
    with pytest.raises(AssertionError) as e:
        sys = starry.System(pri)

    # Different number of wavelength bins
    with pytest.raises(AssertionError) as e:
        sys = starry.System(
            pri, starry.Secondary(starry.Map(ydeg=1, nw=2), porb=1.0)
        )

    # RV for secondary, but not primary
    with pytest.raises(AssertionError) as e:
        sys = starry.System(
            pri, starry.Secondary(starry.Map(rv=True), porb=1.0)
        )

    # Reflected light for first secondary, but not second
    with pytest.raises(ValueError) as e:
        sys = starry.System(
            pri, sec, starry.Secondary(starry.Map(reflected=True), porb=1.0)
        )


def test_bad_sys_data():
    pri = starry.Primary(starry.Map(ydeg=1))
    sec = starry.Secondary(starry.Map(ydeg=1), porb=1.0)
    sys = starry.System(pri, sec)

    # User didn't provide the covariance
    with pytest.raises(ValueError) as e:
        sys.set_data([0.0])

    # User didn't provide a dataset
    with pytest.raises(ValueError) as e:
        sys.solve(t=[0.0])

    # Provide a dummy dataset
    sys.set_data([0.0], C=1.0)

    # User didn't provide a prior for the primary
    with pytest.raises(ValueError) as e:
        sys.solve(t=[0.0])

    # Provide a prior for the primary
    pri.map.set_prior(L=1.0)

    # Provide a prior for the secondary
    sec.map.set_prior(L=1.0)

    # Now check that this works
    sys.solve(t=[0.0])


def test_bad_map_data():
    map = starry.Map(ydeg=1)

    # User didn't provide the covariance
    with pytest.raises(ValueError) as e:
        map.set_data([0.0])

    # User didn't provide a dataset
    with pytest.raises(ValueError) as e:
        map.solve()

    # Provide a dummy dataset
    map.set_data([0.0], C=1.0)

    # User didn't provide a prior
    with pytest.raises(ValueError) as e:
        map.solve()

    # Provide a prior
    map.set_prior(L=1.0)

    # Now check that this works
    map.solve()


def test_bad_map_types():
    # RV + Reflected
    with pytest.raises(NotImplementedError) as e:
        starry.Map(reflected=True, rv=True)

    # Limb-darkened + spectral
    with pytest.raises(NotImplementedError) as e:
        starry.Map(udeg=2, nw=10)



# -*- coding: utf-8 -*-
"""Test map visualization.

These tests don't check for anything; we're just ensuring
the code runs without raising errors.

"""
import starry
import numpy as np
import os
import pymc3 as pm
import pytest


if starry.compat.USE_AESARA:
    theano_config = dict(aesara_config=dict(compute_test_value="ignore"))
else:
    theano_config = dict(theano_config=dict(compute_test_value="ignore"))


# TODO: Several tests segfault on CI. Investigate.
STARRY_TEST_MP4 = False
STARRY_ON_CI = os.getenv("CI", "false") == "true"


def test_show():
    map = starry.Map(ydeg=1, udeg=1)
    map.show(file="tmp.pdf", projection="ortho")
    os.remove("tmp.pdf")
    map.show(file="tmp.pdf", projection="rect")
    os.remove("tmp.pdf")
    if STARRY_TEST_MP4:
        map.show(theta=np.linspace(0, 360, 10), file="tmp.mp4")
        os.remove("tmp.mp4")


def test_show_reflected():
    map = starry.Map(ydeg=1, udeg=1, reflected=True)
    map.show(file="tmp.pdf", projection="ortho")
    os.remove("tmp.pdf")
    map.show(file="tmp.pdf", projection="rect")
    os.remove("tmp.pdf")
    if STARRY_TEST_MP4:
        map.show(theta=np.linspace(0, 360, 10), file="tmp.mp4")
        os.remove("tmp.mp4")


@pytest.mark.skipif(STARRY_ON_CI, reason="Segfaults on CI")
def test_show_rv():
    map = starry.Map(ydeg=1, udeg=1, rv=True)
    map.show(rv=True, file="tmp.pdf", projection="ortho")
    os.remove("tmp.pdf")
    map.show(rv=True, file="tmp.pdf", projection="rect")
    os.remove("tmp.pdf")
    if STARRY_TEST_MP4:
        map.show(rv=True, theta=np.linspace(0, 360, 10), file="tmp.mp4")
        os.remove("tmp.mp4")


def test_show_ld():
    map = starry.Map(udeg=2)
    map.show(file="tmp.pdf")
    os.remove("tmp.pdf")


def test_system_show():
    pri = starry.Primary(starry.Map())
    sec = starry.Secondary(starry.Map(), porb=1.0)
    sys = starry.System(pri, sec)
    sys.show(0.1, file="tmp.pdf")
    os.remove("tmp.pdf")
    sys.show([0.1, 0.2], file="tmp.gif")
    os.remove("tmp.gif")
    if STARRY_TEST_MP4:
        sys.show([0.1, 0.2], file="tmp.mp4")
        os.remove("tmp.mp4")


@pytest.mark.skipif(STARRY_ON_CI, reason="Segfaults on CI")
def test_system_rv_show():
    pri = starry.Primary(starry.Map(rv=True))
    sec = starry.Secondary(starry.Map(rv=True), porb=1.0)
    sys = starry.System(pri, sec)
    sys.show(0.1, file="tmp.pdf")
    os.remove("tmp.pdf")
    if STARRY_TEST_MP4:
        sys.show([0.1, 0.2], file="tmp.mp4")
        os.remove("tmp.mp4")


def test_show_pymc3():
    with pm.Model(**theano_config) as model:
        map = starry.Map()
        ncoeff = map.Ny
        mu = np.ones(ncoeff)
        cov = 1e-2 * np.eye(ncoeff)
        map[:, :] = pm.MvNormal("y", mu, cov, shape=(ncoeff,))
        map.show(theta=0.1, file="tmp.pdf", point=model.test_point)
        os.remove("tmp.pdf")


def test_show_reflected_pymc3():
    with pm.Model(**theano_config) as model:
        map = starry.Map(ydeg=1, udeg=1, reflected=True)
        ncoeff = map.Ny
        mu = np.ones(ncoeff)
        cov = 1e-2 * np.eye(ncoeff)
        map[:, :] = pm.MvNormal("y", mu, cov, shape=(ncoeff,))
        map.show(theta=0.1, file="tmp.pdf", point=model.test_point)
        os.remove("tmp.pdf")


@pytest.mark.skipif(STARRY_ON_CI, reason="Segfaults on CI")
def test_show_rv_pymc3():
    with pm.Model(**theano_config) as model:
        map = starry.Map(ydeg=1, udeg=1, rv=True)
        ncoeff = map.Ny
        mu = np.ones(ncoeff)
        cov = 1e-2 * np.eye(ncoeff)
        map[:, :] = pm.MvNormal("y", mu, cov, shape=(ncoeff,))
        map.show(rv=True, theta=0.1, file="tmp.pdf", point=model.test_point)
        os.remove("tmp.pdf")


def test_show_ld_pymc3():
    with pm.Model(**theano_config) as model:
        map = starry.Map(udeg=2)
        map[1:] = pm.MvNormal("u", [0.5, 0.25], np.eye(2), shape=(2,))
        map.show(file="tmp.pdf", point=model.test_point)
        os.remove("tmp.pdf")


def test_system_show_pymc3():
    with pm.Model(**theano_config) as model:
        pri = starry.Primary(starry.Map())
        sec = starry.Secondary(starry.Map(), porb=1.0)
        sec.inc = pm.Uniform("inc", 0, 90)
        sys = starry.System(pri, sec)
        sys.show(0.1, file="tmp.pdf", point=model.test_point)
        os.remove("tmp.pdf")
        sys.show([0.1, 0.2], file="tmp.gif", point=model.test_point)
        os.remove("tmp.gif")


@pytest.mark.skipif(STARRY_ON_CI, reason="Segfaults on CI")
def test_system_rv_show_pymc3():
    with pm.Model(**theano_config) as model:
        pri = starry.Primary(starry.Map(rv=True))
        sec = starry.Secondary(starry.Map(rv=True), porb=1.0)
        sec.inc = pm.Uniform("inc", 0, 90)
        sys = starry.System(pri, sec)
        sys.show(0.1, file="tmp.pdf", point=model.test_point)
        os.remove("tmp.pdf")



# -*- coding: utf-8 -*-
"""Test Keplerian system stuff."""
import starry
import pytest
from starry.compat import theano
import numpy as np
import astropy.units as u
import exoplanet


def test_compare_to_map_rv():
    """Ensure we get the same result by calling `map.rv()` and `sys.rv()`.
    """
    # Define the map
    map = starry.Map(ydeg=1, udeg=2, rv=True, amp=1, veq=1, alpha=0)
    map[1, 0] = 0.5

    # Define the star
    A = starry.Primary(
        map, r=1.0, m=1.0, prot=0, length_unit=u.Rsun, mass_unit=u.Msun
    )

    # Define the planet
    b = starry.Secondary(
        starry.Map(rv=True, amp=1, veq=0),
        r=0.1,
        porb=1.0,
        m=0.01,
        t0=0.0,
        inc=86.0,
        ecc=0.3,
        w=60,
        length_unit=u.Rsun,
        mass_unit=u.Msun,
        angle_unit=u.degree,
        time_unit=u.day,
    )

    # Define the system
    sys = starry.System(A, b)

    # Time array
    time = np.linspace(-0.05, 0.05, 1000)

    # Get the positions of both bodies
    x, y, z = sys.position(time)

    # Compute the relative positions
    xo = x[1] - x[0]
    yo = y[1] - y[0]
    zo = z[1] - z[0]
    ro = b.r / A.r

    # Compare
    rv1 = map.rv(xo=xo, yo=yo, ro=ro)
    rv2 = sys.rv(time, keplerian=False)
    assert np.allclose(rv1, rv2)


def test_compare_to_exoplanet():
    """Ensure we get the same result with `starry` and `exoplanet`.
    """
    # Define the star
    A = starry.Primary(
        starry.Map(rv=True, veq=0),
        r=1.0,
        m=1.0,
        prot=0,
        length_unit=u.Rsun,
        mass_unit=u.Msun,
    )

    # Define the planet
    b = starry.Secondary(
        starry.Map(rv=True, veq=0),
        r=0.1,
        porb=1.0,
        m=0.01,
        t0=0.0,
        inc=86.0,
        ecc=0.3,
        w=60,
        length_unit=u.Rsun,
        mass_unit=u.Msun,
        angle_unit=u.degree,
        time_unit=u.day,
    )

    # Define the planet
    c = starry.Secondary(
        starry.Map(rv=True, veq=0),
        r=0.1,
        porb=1.7,
        m=0.02,
        t0=0.3,
        inc=87.0,
        ecc=0.2,
        w=70,
        length_unit=u.Rsun,
        mass_unit=u.Msun,
        angle_unit=u.degree,
        time_unit=u.day,
    )

    # Define the system
    sys = starry.System(A, b, c)

    # Compute with starry
    time = np.linspace(-0.5, 0.5, 1000)
    rv1 = sys.rv(time, keplerian=True, total=True)

    # Compute with exoplanet
    orbit = exoplanet.orbits.KeplerianOrbit(
        period=[1.0, 1.7],
        t0=[0.0, 0.3],
        incl=[86.0 * np.pi / 180, 87.0 * np.pi / 180],
        ecc=[0.3, 0.2],
        omega=[60 * np.pi / 180, 70 * np.pi / 180],
        m_planet=[0.01, 0.02],
        m_star=1.0,
        r_star=1.0,
    )
    rv2 = orbit.get_radial_velocity(time).eval().sum(axis=1)

    assert np.allclose(rv1, rv2)



# -*- coding: utf-8 -*-
"""Test the map intensity evaluation."""
import starry
import numpy as np
import pytest

# Setup
map = starry.Map(ydeg=1)
params = [
    [[0, 0, 0], 0, 0, 1.0 / np.pi],
    [[1, 0, 0], 0, np.linspace(-180, 180, 10), 1.0 / np.pi],
    [[0, 1, 0], 0, 0, 1.0 / np.pi + np.sqrt(3) / np.pi],
    [[0, 0, 1], np.linspace(-90, 90, 10), 0, 1.0 / np.pi],
    [
        [1, 0, 0],
        30,
        np.linspace(-180, 180, 10),
        1.0 / np.pi + np.sqrt(3) / (2.0 * np.pi),
    ],
]


@pytest.mark.parametrize("y,lat,lon,I", params)
def test_intensity(y, lat, lon, I):
    map[1:, :] = y
    assert np.allclose(map.intensity(lat=lat, lon=lon), I)


def test_design_matrix():
    map[1:, :] = [0.1, 0.2, 0.3]
    lat = [0, 30]
    lon = [45, 90]
    P = map.intensity_design_matrix(lat=lat, lon=lon)
    assert np.allclose(map.intensity(lat=lat, lon=lon), P.dot(map.y))


def test_limb_darkened():
    map = starry.Map(udeg=2)
    map[1] = 0.5
    map[2] = 0.25

    assert map.intensity(mu=0) == 1 - map[1] - map[2]  # limb
    assert map.intensity(mu=1) == 1.0  # center

    assert map.intensity(x=1, y=0) == 1 - map[1] - map[2]  # limb
    assert map.intensity(x=0, y=0) == 1.0  # center



from starry.compat import theano
from starry.compat import tt
import numpy as np
import starry
import matplotlib.pyplot as plt
import pytest


@pytest.fixture
def model():
    class Model:
        def __init__(self):
            self.map = starry.Map(ydeg=1, reflected=True)
            _b = tt.dvector("b")
            _theta = tt.dvector("theta")
            _bo = tt.dvector("bo")
            _ro = tt.dscalar("ro")
            _sigr = tt.dscalar("sigr")
            self._s = theano.function(
                [_b, _theta, _bo, _ro, _sigr],
                self.map.ops.sT(_b, _theta, _bo, _ro, _sigr),
            )

            self._dsdb = [
                theano.function(
                    [_b, _theta, _bo, _ro, _sigr],
                    tt.grad(
                        self.map.ops.sT(_b, _theta, _bo, _ro, _sigr)[0, n], _b
                    ),
                )
                for n in range(self.map.Ny)
            ]

            self._dsdtheta = [
                theano.function(
                    [_b, _theta, _bo, _ro, _sigr],
                    tt.grad(
                        self.map.ops.sT(_b, _theta, _bo, _ro, _sigr)[0, n],
                        _theta,
                    ),
                )
                for n in range(self.map.Ny)
            ]

            self._dsdbo = [
                theano.function(
                    [_b, _theta, _bo, _ro, _sigr],
                    tt.grad(
                        self.map.ops.sT(_b, _theta, _bo, _ro, _sigr)[0, n], _bo
                    ),
                )
                for n in range(self.map.Ny)
            ]

            self._dsdro = [
                theano.function(
                    [_b, _theta, _bo, _ro, _sigr],
                    tt.grad(
                        self.map.ops.sT(_b, _theta, _bo, _ro, _sigr)[0, n], _ro
                    ),
                )
                for n in range(self.map.Ny)
            ]

            self._dsdsigr = [
                theano.function(
                    [_b, _theta, _bo, _ro, _sigr],
                    tt.grad(
                        self.map.ops.sT(_b, _theta, _bo, _ro, _sigr)[0, n],
                        _sigr,
                    ),
                )
                for n in range(self.map.Ny)
            ]

        def s(self, b, theta, bo, ro, sigr, n=0):
            if hasattr(ro, "__len__"):
                assert not (
                    hasattr(b, "__len__")
                    or hasattr(theta, "__len__")
                    or hasattr(bo, "__len__")
                    or hasattr(sigr, "__len__")
                )
                return [
                    self._s([b], [theta], [bo], ro[i], sigr)[0, n]
                    for i in range(len(ro))
                ]
            elif hasattr(sigr, "__len__"):
                assert not (
                    hasattr(b, "__len__")
                    or hasattr(theta, "__len__")
                    or hasattr(bo, "__len__")
                    or hasattr(ro, "__len__")
                )
                return [
                    self._s([b], [theta], [bo], ro, sigr[i])[0, n]
                    for i in range(len(sigr))
                ]
            else:
                assert (
                    hasattr(b, "__len__")
                    or hasattr(theta, "__len__")
                    or hasattr(bo, "__len__")
                )
                shaper = (
                    np.zeros_like(b) + np.zeros_like(bo) + np.zeros_like(theta)
                )
                b += shaper
                theta += shaper
                bo += shaper
                return self._s(b, theta, bo, ro, sigr)[:, n]

        def dsdb(self, b, theta, bo, ro, sigr, n=0):
            b = np.atleast_1d(b)
            assert not hasattr(theta, "__len__")
            assert not hasattr(bo, "__len__")
            assert not hasattr(ro, "__len__")
            assert not hasattr(sigr, "__len__")
            return np.array(
                [
                    self._dsdb[n]([b[i]], [theta], [bo], ro, sigr)
                    for i in range(len(b))
                ]
            )

        def dsdtheta(self, b, theta, bo, ro, sigr, n=0):
            theta = np.atleast_1d(theta)
            assert not hasattr(b, "__len__")
            assert not hasattr(bo, "__len__")
            assert not hasattr(ro, "__len__")
            assert not hasattr(sigr, "__len__")
            return np.array(
                [
                    self._dsdtheta[n]([b], [theta[i]], [bo], ro, sigr)
                    for i in range(len(theta))
                ]
            )

        def dsdbo(self, b, theta, bo, ro, sigr, n=0):
            bo = np.atleast_1d(bo)
            assert not hasattr(b, "__len__")
            assert not hasattr(theta, "__len__")
            assert not hasattr(ro, "__len__")
            assert not hasattr(sigr, "__len__")
            return np.array(
                [
                    self._dsdbo[n]([b], [theta], [bo[i]], ro, sigr)
                    for i in range(len(bo))
                ]
            )

        def dsdro(self, b, theta, bo, ro, sigr, n=0):
            ro = np.atleast_1d(ro)
            assert not hasattr(b, "__len__")
            assert not hasattr(theta, "__len__")
            assert not hasattr(bo, "__len__")
            assert not hasattr(sigr, "__len__")
            return np.array(
                [
                    self._dsdro[n]([b], [theta], [bo], ro[i], sigr)
                    for i in range(len(ro))
                ]
            )

        def dsdsigr(self, b, theta, bo, ro, sigr, n=0):
            sigr = np.atleast_1d(sigr)
            assert not hasattr(b, "__len__")
            assert not hasattr(theta, "__len__")
            assert not hasattr(bo, "__len__")
            assert not hasattr(ro, "__len__")
            return np.array(
                [
                    self._dsdsigr[n]([b], [theta], [bo], ro, sigr[i])
                    for i in range(len(sigr))
                ]
            )

        def grad(self, b, theta, bo, ro, sigr, n=0):
            if hasattr(b, "__len__"):
                wrt = b
            elif hasattr(theta, "__len__"):
                wrt = theta
            elif hasattr(bo, "__len__"):
                wrt = bo
            elif hasattr(ro, "__len__"):
                wrt = ro
            elif hasattr(sigr, "__len__"):
                wrt = sigr
            else:
                assert False
            return np.gradient(
                self.s(b, theta, bo, ro, sigr, n=n), edge_order=2
            ) / np.gradient(wrt, edge_order=2)

    return Model()


def test_derivs(model, n=1, npts=10000, atol=1e-5, plot=False, throw=True):

    if plot:
        fig, ax = plt.subplots(3, 5, figsize=(16, 7))

    # b gradient
    theta = 0.51
    bo = 0.75
    ro = 0.1
    sigr = 0.0
    b = np.linspace(-1, 1, npts)
    g1 = model.grad(b, theta, bo, ro, sigr, n=n)
    g2 = model.dsdb(b, theta, bo, ro, sigr, n=n).flatten()
    # Pad the edges (numerical gradient isn't great)
    if throw:
        assert np.allclose(g1[200:-50], g2[200:-50], atol=atol), "error in b"

    if plot:
        ax[0, 0].plot(b, s(b, theta, bo, ro, sigr, n=n))
        ax[1, 0].plot(b, g1, lw=2)
        ax[1, 0].plot(b, g2, lw=1)
        ax[2, 0].plot(b, np.log10(np.abs(g1 - g2)), "k")
        ax[2, 0].set_xlabel("b")
        for axis in ax[:, 0]:
            axis.set_xlim(-1, 1)

    # theta gradient
    b = 0.51
    bo = 0.75
    ro = 0.1
    sigr = 30 * np.pi / 180
    theta = np.linspace(-np.pi, np.pi, npts)
    g1 = model.grad(b, theta, bo, ro, sigr, n=n)
    g2 = model.dsdtheta(b, theta, bo, ro, sigr, n=n).flatten()
    if throw:
        assert np.allclose(g1, g2, atol=atol), "error in theta"

    if plot:
        ax[0, 1].plot(theta, s(b, theta, bo, ro, sigr, n=n))
        ax[1, 1].plot(theta, g1, lw=2)
        ax[1, 1].plot(theta, g2, lw=1)
        ax[2, 1].plot(theta, np.log10(np.abs(g1 - g2)), "k")
        ax[2, 1].set_xlabel("theta")
        for axis in ax[:, 1]:
            axis.set_xlim(-np.pi, np.pi)

    # bo gradient
    b = 0.51
    theta = 0.49
    ro = 0.1
    sigr = 30 * np.pi / 180
    bo = np.linspace(0, 1.5, npts)
    g1 = model.grad(b, theta, bo, ro, sigr, n=n)
    g2 = model.dsdbo(b, theta, bo, ro, sigr, n=n).flatten()
    if throw:
        assert np.allclose(g1, g2, atol=atol), "error in bo"

    if plot:
        ax[0, 2].plot(bo, s(b, theta, bo, ro, sigr, n=n))
        ax[1, 2].plot(bo, g1, lw=2)
        ax[1, 2].plot(bo, g2, lw=1)
        ax[2, 2].plot(bo, np.log10(np.abs(g1 - g2)), "k")
        ax[2, 2].set_xlabel("bo")
        for axis in ax[:, 2]:
            axis.set_xlim(0, 1.5)

    # ro gradient
    b = 0.51
    theta = 0.49
    bo = 0.75
    sigr = 30 * np.pi / 180
    ro = np.linspace(0.001, 1.5, npts)
    g1 = model.grad(b, theta, bo, ro, sigr, n=n)
    g2 = model.dsdro(b, theta, bo, ro, sigr, n=n).flatten()
    if throw:
        assert np.allclose(g1, g2, atol=atol), "error in ro"

    if plot:
        ax[0, 3].plot(ro, s(b, theta, bo, ro, sigr, n=n))
        ax[1, 3].plot(ro, g1, lw=2)
        ax[1, 3].plot(ro, g2, lw=1)
        ax[1, 3].set_ylim(-1.5, 1.0)
        ax[2, 3].plot(ro, np.log10(np.abs(g1 - g2)), "k")
        ax[2, 3].set_xlabel("ro")
        for axis in ax[:, 3]:
            axis.set_xlim(0.001, 1.5)

        for axis in ax[2, :]:
            axis.set_ylim(-6, 0)

    # sigr gradient
    b = 0.51
    theta = 0.49
    bo = 0.75
    ro = 0.1
    sigr = np.linspace(0, 30 * np.pi / 180, npts)
    g1 = model.grad(b, theta, bo, ro, sigr, n=n)
    g2 = model.dsdsigr(b, theta, bo, ro, sigr, n=n).flatten()
    if throw:
        assert np.allclose(g1, g2, atol=atol), "error in sigr"

    if plot:
        ax[0, 4].plot(sigr, s(b, theta, bo, ro, sigr, n=n))
        ax[1, 4].plot(sigr, g1, lw=2)
        ax[1, 4].plot(sigr, g2, lw=1)
        ax[1, 4].set_ylim(-1.5, 1.0)
        ax[2, 4].plot(sigr, np.log10(np.abs(g1 - g2)), "k")
        ax[2, 4].set_xlabel("sigr")
        for axis in ax[:, 4]:
            axis.set_xlim(0, 30 * np.pi / 180)

        plt.show()


def test_abs_b_one():
    """Check derivs are finite when b=+/-1."""
    b = tt.dscalar()
    b.tag.test_value = -1.0
    map = starry.Map(reflected=True)

    def flux(b):
        return map.flux(zs=-b, ys=0)

    grad = theano.function([b], tt.grad(flux(b)[0], [b]))
    assert not np.isnan(grad(-1.0)[0]) and not np.isnan(grad(1.0)[0])


if __name__ == "__main__":
    test_abs_b_one()



"""
Extension for querying NExSci for real planet systems.

*Created by* `Christina Hedges <mailto:christina.l.hedges@nasa.gov>`_ *on
November 1, 2019*

Example
-------
Obtain the :py:class:`starry.System` for Kepler-10 and evaluate at
some times. Here ``flux`` contains the normalized flux from the
Kepler-10 system.

.. code-block:: python

    from starry.extensions import from_nexsci
    import numpy as np
    sys = from_nexsci('Kepler-10')
    time = np.arange(0, 10, 0.1)
    flux = sys.flux(time).eval()

"""
import pandas as pd
import numpy as np
import os
import datetime
import astropy.units as u
from glob import glob
import logging

from ... import _PACKAGEDIR
from ... import Secondary, Primary, System, Map


logger = logging.getLogger("starry.nexsci")


class FromNexsci:
    """
    Extension for retrieving a :py:class:`starry.System` initialized
    with parameters from NExSci.

    Specify the name of the system as a string to ``name`` to create that specific
    system. Users can also pass ``'random'`` to obtain a random system. Returns
    a :py:class:`starry.System` object.

    This extension queries NExSci and then stores the data as a csv file in
    ``/extensions/data``. Loading this module will trigger a query of NExSci if
    this csv file is more than 7 days out of date.

    Args:
        name (str): Name of system to retrieve.
            e.g. 'K2-43' or 'K2-43b'
        limb_darkening (list, optional): The quadratic limb-darkening
            parameters to apply to the model. Default is ``[0.4, 0.2]``
    Returns:
        An instance of :py:class:`starry.System` using planet parameters \
        from NExSci
    """

    def __init__(self):
        fname = glob(
            "{}/extensions/nexsci/data/planets.csv".format(_PACKAGEDIR)
        )
        if len(fname) == 0:
            self.stale = True
        else:
            st = os.stat(fname[0])
            mtime = st.st_mtime
            # If database is out of date, get it again.
            if datetime.datetime.now() - datetime.datetime.fromtimestamp(
                mtime
            ) > datetime.timedelta(days=7):
                self.stale = True
            else:
                self.stale = False

    def __call__(self, name, limb_darkening=[0.4, 0.2]):

        df = self._get_nexsci_data()
        if name == "random":
            system_list = (
                df[
                    [
                        "pl_hostname",
                        "pl_letter",
                        "pl_orbper",
                        "pl_trandep",
                        "pl_radj",
                    ]
                ]
                .dropna()["pl_hostname"]
                .unique()
            )
            name = np.random.choice(system_list)
            logger.warning("Randomly selected {}".format(name))
            df = df[df.pl_hostname == name].reset_index(drop=True)
        else:
            sname = name.lower().replace("-", "").replace(" ", "")
            if np.any([name.endswith(i) for i in "bcdefghijklmnopqrstuvwxyz"]):
                df = df[
                    (df.stripped_name + df.pl_letter) == sname
                ].reset_index(drop=True)
            else:
                df = df[(df.stripped_name) == sname].reset_index(drop=True)
            if len(df) == 0:
                raise ValueError("{} does not exist at NExSci".format(name))

        m = df.iloc[0].st_mass
        if ~np.isfinite(m):
            m = 1
            logger.warning("Stellar mass is NaN, setting to 1.")

        star = Primary(
            Map(udeg=len(limb_darkening), amp=1),
            r=np.asarray(df.iloc[0]["st_rad"]),
            m=m,
        )
        for idx in range(len(limb_darkening)):
            star.map[idx + 1] = limb_darkening[idx]
        planets = []
        for idx, d in df.iterrows():
            for key in ["pl_orbper", "pl_tranmid", "pl_radj"]:
                if ~np.isfinite(d[key]):
                    logger.warning(
                        "{} is not set for {}{}, planet will be skipped".format(
                            key, d.pl_hostname, d.pl_letter
                        )
                    )
            r = np.asarray(d.pl_radj * u.jupiterRad.to(u.solRad))
            planet = Secondary(
                Map(amp=0),
                porb=d.pl_orbper,
                t0=d.pl_tranmid,
                r=r,
                inc=d.pl_orbincl,
                ecc=d.pl_eccen,
                w=d.pl_orblper,
            )
            planets.append(planet)
        sys = System(star, *planets)
        return sys

    def _fill_data(self, df):
        """Takes a dataframe of nexsci data and fills in NaNs with sane defaults.

        Parameters
        ----------
        df : pandas.DataFrame
            Dataframe of planet parameters from nexsci

        Returns
        -------
        clean_df : pandas.DataFrame
            Dataframe of planet parameters from nexsci, with NaNs replaced.
        """

        # Supplemental data for some planets
        sup = pd.read_csv(
            "{}/extensions/nexsci/data/supplement.csv".format(_PACKAGEDIR)
        )
        for label in [
            "pl_orbincl",
            "pl_eccen",
            "pl_trandep",
            "pl_tranmid",
            "pl_orbper",
        ]:
            nan = ~np.isfinite(df[label])
            vals = np.asarray(
                df[nan][["stripped_name", "pl_letter"]].merge(
                    sup[["stripped_name", "pl_letter", label]], how="left"
                )[label]
            )
            df.loc[nan, label] = vals

        nan = ~np.isfinite(df.pl_orbincl)
        df.loc[nan, ["pl_orbincl"]] = np.zeros(nan.sum()) + 90

        nan = ~np.isfinite(df.pl_eccen)
        df.loc[nan, ["pl_eccen"]] = np.zeros(nan.sum())

        nan = ~np.isfinite(df.pl_orblper)
        df.loc[nan, ["pl_orblper"]] = np.zeros(nan.sum())

        # Fill in missing trandep
        nan = ~np.isfinite(df.pl_trandep)
        trandep = (
            np.asarray(df.pl_radj * u.jupiterRad.to(u.solRad))
            / np.asarray(df.st_rad)
        ) ** 2
        df.loc[nan, ["pl_trandep"]] = trandep[nan]
        return df

    def _retrieve_online_data(self, guess_masses=False):
        """Queries nexsci database and returns a dataframe of planet data.
        """
        logger.warning("Database out of date. Redownloading...")
        NEXSCI_API = "http://exoplanetarchive.ipac.caltech.edu/cgi-bin/nstedAPI/nph-nstedAPI"
        try:
            # Planet table
            planets = pd.read_csv(
                NEXSCI_API + "?table=planets&select=pl_hostname,pl_letter,"
                "pl_disc,ra,dec,pl_trandep,pl_tranmid,pl_tranmiderr1,"
                "pl_tranmiderr2,pl_tranflag,pl_trandur,pl_pnum,pl_k2flag,"
                "pl_kepflag,pl_facility,pl_orbincl,pl_orbinclerr1,pl_orbinclerr2,"
                "pl_orblper,st_mass,st_masserr1,st_masserr2,st_rad,st_raderr1,"
                "st_raderr2,st_teff,st_tefferr1,st_tefferr2,st_optmag,st_j,st_h",
                comment="#",
                skiprows=1,
            )
            # Composite table
            composite = pd.read_csv(
                NEXSCI_API
                + "?table=compositepars&select=fpl_hostname,fpl_letter,"
                "fpl_smax,fpl_smaxerr1,fpl_smaxerr2,fpl_radj,fpl_radjerr1,"
                "fpl_radjerr2,fpl_bmassj,fpl_bmassjerr1,fpl_bmassjerr2,"
                "fpl_bmassprov,fpl_eqt,fpl_orbper,fpl_orbpererr1,fpl_orbpererr2,"
                "fpl_eccen,fpl_eccenerr1,fpl_eccenerr2,fst_spt",
                comment="#",
                skiprows=1,
            )
            # Rename columns
            composite.columns = [
                "pl_hostname",
                "pl_letter",
                "pl_orbsmax",
                "pl_orbsmaxerr1",
                "pl_orbsmaxerr2",
                "pl_radj",
                "pl_radjerr1",
                "pl_radjerr2",
                "pl_bmassj",
                "pl_bmassjerr1",
                "pl_bmassjerr2",
                "pl_bmassprov",
                "pl_eqt",
                "pl_orbper",
                "pl_orbpererr1",
                "pl_orbpererr2",
                "pl_eccen",
                "pl_eccenerr1",
                "pl_eccenerr2",
                "st_spt",
            ]
        except Exception as e:
            logger.warning(
                "Couldn't obtain data from NEXSCI. Do you have an internet connection?"
            )
            return
        df = pd.merge(
            left=planets,
            right=composite,
            how="left",
            left_on=["pl_hostname", "pl_letter"],
            right_on=["pl_hostname", "pl_letter"],
        )

        df["stripped_name"] = np.asarray(
            [
                n.lower().replace("-", "").replace(" ", "")
                for n in df.pl_hostname
            ]
        )
        df = df[df.pl_tranflag == 1].reset_index(drop=True)

        df = self._fill_data(df)

        df = df.to_csv(
            "{}/extensions/nexsci/data/planets.csv".format(_PACKAGEDIR),
            index=False,
        )

        self.stale = False

    def _get_nexsci_data(self):
        """Returns pandas dataframe of all exoplanet data from nexsci
        """
        if self.stale:
            self._retrieve_online_data()
        return pd.read_csv(
            "{}/extensions/nexsci/data/planets.csv".format(_PACKAGEDIR)
        )


from_nexsci = FromNexsci()



# -*- coding: utf-8 -*-
import warnings
import aesara_theano_fallback
from aesara_theano_fallback import aesara as theano
import aesara_theano_fallback.tensor as tt
from aesara_theano_fallback import sparse as ts
from aesara_theano_fallback import change_flags, ifelse, USE_AESARA
from aesara_theano_fallback.tensor import slinalg
from aesara_theano_fallback.graph import basic, op, params_type, fg

__all__ = [
    "theano",
    "tt",
    "ts",
    "slinalg",
    "ifelse",
    "Apply",
    "COp",
    "Op",
    "Params",
    "ParamsType",
    "Node",
    "change_flags",
    "floatX",
    "evaluator",
    "USE_AESARA",
]

# Suppress third-party deprecation warnings
warnings.filterwarnings("ignore", category=DeprecationWarning, module="pymc3")
warnings.filterwarnings("ignore", category=DeprecationWarning, module="theano")
warnings.filterwarnings("ignore", category=DeprecationWarning, module="aesara")

# Set double precision
floatX = "float64"

# Compatibility imports
Node = basic.Node
Apply = basic.Apply
Op = op.Op
COp = op.ExternalCOp
Params = params_type.Params
ParamsType = params_type.ParamsType
MissingInputError = fg.MissingInputError
theano.config.floatX = floatX


def evaluator(**kwargs):
    """
    Return a function to evaluate theano tensors.

    Works inside a `pymc3` model if a `point` is provided.
    Lazily imports `pymc3` to minimize overhead.

    """
    # Store the kwargs
    kwargs_point = kwargs.pop("point", None)
    kwargs_model = kwargs.pop("model", None)

    if kwargs_point is not None:

        # User provided a point

        import pymc3 as pm
        import pymc3_ext as pmx

        point = kwargs_point
        model = kwargs_model
        if model is None:
            model = pm.Model.get_context()
        get_val = lambda x: pmx.eval_in_model(x, model=model, point=point)

    else:

        # No point provided

        def get_val(x):

            try:

                # Try to directly evaluate it

                return x.eval()

            except MissingInputError as e:

                # That didn't work. Perhaps we are in a pymc3 model
                # context, but the user didn't provide a point?

                import pymc3 as pm
                import pymc3_ext as pmx

                try:
                    model = kwargs_model
                    if model is None:
                        model = pm.Model.get_context()
                except TypeError:
                    raise ValueError(
                        "Missing input for variable {}, and no pymc3 model found.".format(
                            x
                        )
                    )

                # Warn the user that we're using the test point
                warnings.warn(
                    "Detected pymc3 model context, but no point provided. Evaluating at test_point."
                )

                return pmx.eval_in_model(
                    x, model=model, point=model.test_point
                )

    return get_val



# -*- coding: utf-8 -*-
import numpy as np


__all__ = ["get_ylm_inds", "get_ul_inds", "get_ylmw_inds", "integers"]


integers = (int, np.int16, np.int32, np.int64)


def get_ylm_inds(ydeg, ls, ms):
    """

    """

    # Turn `ls` and `ms` into slices
    if isinstance(ls, integers):
        ls = slice(ls, ls + 1)
    if isinstance(ms, integers):
        ms = slice(ms, ms + 1)

    if isinstance(ls, slice) and isinstance(ms, slice):

        # List of indices user is accessing
        inds = []

        # Fill in the `None`s
        if ls.start is None:
            ls = slice(0, ls.stop, ls.step)
        if ls.stop is None:
            ls = slice(ls.start, ydeg + 1, ls.step)
        if ls.step is None:
            ls = slice(ls.start, ls.stop, 1)
        if ms.step is None:
            ms = slice(ms.start, ms.stop, 1)

        if (ls.start < 0) or (ls.start > ydeg):
            raise ValueError("Invalid value for `l`.")

        # Loop through all the Ylms
        for l in range(ls.start, ls.stop, ls.step):
            ms_ = slice(ms.start, ms.stop, ms.step)
            if (ms_.start is None) or (ms_.start < -l):
                ms_ = slice(-l, ms_.stop, ms_.step)
            if (ms_.stop is None) or (ms_.stop > l):
                ms_ = slice(ms_.start, l + 1, ms_.step)
            for m in range(ms_.start, ms_.stop, ms_.step):
                n = l * l + l + m
                if (
                    (n < 0) or (n >= (ydeg + 1) ** 2) or (m > l) or (m < -l)
                ):  # pragma: no cover
                    raise ValueError("Invalid value for `l` and/or `m`.")
                inds.append(n)

        return inds

    else:

        # Not a slice, not an int... What is it?
        raise ValueError("Invalid value for `l` and/or `m`.")


def get_ylmw_inds(ydeg, nw, ls, ms, ws):
    """

    """

    # Turn the `ws` into slices
    if isinstance(ws, integers):
        ws = slice(ws, ws + 1)

    if isinstance(ws, slice):

        # Get the `l`, `m` indices
        lminds = get_ylm_inds(ydeg, ls, ms)

        # Process the `w` indices
        winds = []

        # Fill in the `None`s
        if ws.start is None:
            ws = slice(0, ws.stop, ws.step)
        if ws.stop is None:
            ws = slice(ws.start, nw, ws.step)
        if ws.step is None:
            ws = slice(ws.start, ws.stop, 1)

        if (ws.start < 0) or (ws.start >= nw):
            raise ValueError("Invalid value for `w`.")

        return tuple((lminds, ws))

    else:

        # Not a slice, not an int... What is it?
        raise ValueError("Invalid value for `w`.")


def get_ul_inds(udeg, ls):
    """

    """

    # Turn `ls` into a slice
    if isinstance(ls, integers):
        ls = slice(ls, ls + 1)

    if isinstance(ls, slice):

        # List of indices user is accessing
        inds = []

        # Fill in the `None`s
        if ls.start is None:
            ls = slice(0, ls.stop, ls.step)
        if ls.stop is None:
            ls = slice(ls.start, udeg + 1, ls.step)
        if ls.step is None:
            ls = slice(ls.start, ls.stop, 1)

        if (ls.start < 0) or (ls.start > udeg):
            raise ValueError("Invalid value for `l`.")

        # Loop through all the `ls`
        for l in range(ls.start, ls.stop, ls.step):
            inds.append(l)

        return inds

    else:  # pragma: no cover

        # Not a slice, not an int... What is it?
        raise ValueError("Invalid value for `l`.")



import starry
import numpy as np
import matplotlib.pyplot as plt
import pytest


@pytest.mark.parametrize("noon", [False, True])
def test_I_stability(noon, plot=False):
    # FB found this unstable limit. The instability comes
    # from two places:
    # 1. The upward recursion in the I integral is unstable.
    #    We need to either implement a tridiagonal solver as
    #    in the J integral and/or refine our computation of kappa.
    # 2. The terminator parameter b = 1, which causes `get_angles` to
    #    oscillate between different integration codes. When b
    #    approaches unity, we should switch to the regular starry
    #    solver, since the terminator is so close to the limb that its
    #    presence doesn't matter.
    xo, yo, zo, ro = (
        31.03953239062832,
        23.892679948795926,
        1.0,
        39.10406741663172,
    )
    # Exactly noon?
    if noon:
        xs = 0.0
    else:
        xs = 0.1
    ys = 0.0
    zs = 1.0
    xo = np.linspace(xo - 2, xo + 2, 1000)

    # Compute analytic
    map = starry.Map(ydeg=10, reflected=True)
    map[10, :] = 1
    flux1 = map.flux(xo=xo, yo=yo, zo=zo, ro=ro, xs=xs, ys=ys, zs=zs)

    if noon:
        # The flux above should be *exactly* equal to 2/3 the flux of a
        # linearly-limb darkened source with u_1 = 1.0, since linear
        # limb darkening weights the surface brightness by the same
        # cosine-like profile (2/3 is the geometrical albedo of a
        # perfect Lambert sphere)
        map_e = starry.Map(ydeg=10, udeg=1)
        map_e[10, :] = 1
        map_e[1] = 1
        flux2 = (2.0 / 3.0) * map_e.flux(xo=xo, yo=yo, zo=zo, ro=ro)
        atol = 1e-12

    else:

        # Compute numerical
        res = 500
        (lat, lon), (x, y, z) = map.ops.compute_ortho_grid(res)
        image = map.render(xs=xs, ys=ys, zs=zs, res=res).flatten()
        flux2 = np.zeros_like(flux1)
        for k in range(len(xo)):
            idx = (x - xo[k]) ** 2 + (y - yo) ** 2 > ro ** 2
            flux2[k] = np.nansum(image[idx])
        flux2 *= 4 / res ** 2
        atol = 1e-3

    # Plot it
    if plot:
        plt.plot(xo, flux1)
        plt.plot(xo, flux2)
        plt.show()

    # Compare
    assert np.allclose(flux1, flux2, atol=atol)


if __name__ == "__main__":
    starry.config.lazy = False
    # test_I_stability(False, plot=True)
    test_I_stability(True, plot=True)



"""
Starry was originally designed for maps in thermal light.
Maps in reflected light are quirky in how things are normalized,
so we need to make sure albedos, intensities, and fluxes are
all self-consistent in the code. There are some fudges we had
to make under the hood to ensure all of these tests pass.

"""

import numpy as np
import starry


starry.config.lazy = False
map = starry.Map(1, reflected=True)


def test_amp():
    assert map.amp == 1.0


def test_flux():
    assert np.allclose(map.flux(zs=1, xs=0, ys=0), 2.0 / 3.0)


def test_design_matrix():
    assert np.allclose(
        map.amp * map.design_matrix(zs=1, xs=0, ys=0)[0, 0], 2.0 / 3.0
    )


def test_intensity():
    assert np.allclose(
        map.intensity(lat=0, lon=0, zs=1, xs=0, ys=0), 1.0 / np.pi
    )


def test_albedo():
    assert np.allclose(
        map.amp
        * map.intensity(lat=0, lon=0, zs=1, xs=0, ys=0, illuminate=False),
        1.0,
    )


def test_albedo_design_matrix():
    assert np.allclose(
        map.amp * map.intensity_design_matrix(lat=0, lon=0)[0, 0], 1.0
    )


def test_render():
    assert np.allclose(np.nanmax(map.render(zs=1, xs=0, ys=0)), 1.0 / np.pi)


def test_render_albedo():
    assert np.allclose(
        np.nanmax(map.render(zs=1, xs=0, ys=0, illuminate=False)), 1.0
    )


def test_numerical_flux(res=300):
    res = 300
    atol = 1e-4
    rtol = 1e-4
    num_flux = np.nansum(map.render(zs=1, xs=0, ys=0, res=res)) * 4 / res ** 2
    assert np.allclose(num_flux, 2.0 / 3.0, atol=atol, rtol=rtol)


def test_solve():
    # Generate a light curve with two points: one at
    # full phase, one at new phase.
    np.random.seed(0)
    ferr = 1e-8
    flux = np.array([2.0 / 3.0, 0])
    flux += ferr * np.random.randn(len(flux))
    map.set_data(flux, C=ferr ** 2)

    # Make the prior very informative for the coeffs
    # but extremely wide for the amplitude (with zero mean)
    mu = np.zeros(4)
    L = np.array([1e1, 1e-12, 1e-12, 1e-12])
    map.set_prior(mu=mu, L=L)

    # Check that we recover the correct amplitude
    x, _ = map.solve(zs=[1.0, -1.0])
    assert np.allclose(x[0], 1.0)
    assert np.allclose(map.amp, 1.0)

    # Check that a map draw gives us the correct amplitude
    map.draw()
    assert np.allclose(map.amp, 1.0)


def test_one_over_r_squared(n_tests=10, plot=False):
    """Test that the flux decreases as 1/r^2."""
    flux0 = map.flux()
    zs = np.linspace(1, 10, 100)
    flux = map.flux(xs=0, ys=0, zs=zs)
    assert np.allclose(flux, flux0 / zs ** 2)


def test_sys_flux():
    """Test the normalization of the flux."""
    # Instantiate a system. Planet has radius `r` and is at
    # distance `d` from a point illumination source.
    d = 10
    r = 2
    planet = starry.Secondary(starry.Map(reflected=True), a=d, r=r)
    star = starry.Primary(starry.Map(), r=0)
    sys = starry.System(star, planet)

    # Get the star & planet flux when it's at full phase
    t_full = 0.5 * sys._get_periods()[0]
    f_star, f_planet = sys.flux(t=t_full, total=False)

    # Star should have unit flux
    assert np.allclose(f_star, 1.0)

    # Planet should have flux equal to (2 / 3) r^2 / d^2
    f_expected = (2.0 / 3.0) * r ** 2 / d ** 2
    assert np.allclose(f_planet, f_expected)



from starry._c_ops import Ops
import pickle
import numpy as np


def test_pickle():
    ops = Ops(5, 0, 0)
    data = pickle.dumps(ops, -1)
    new_ops = pickle.loads(data)
    assert np.allclose(ops.rTA1, new_ops.rTA1)



# -*- coding: utf-8 -*-
"""
Test the nexsci extension.

"""
import numpy as np
import pandas as pd
import pytest
import starry
import warnings
from starry.extensions import from_nexsci

starry.config.lazy = True


def test_lazy_nexsci_query():
    """ Tests if the nexsci query works. """

    # These should run without error
    from_nexsci._retrieve_online_data()

    df = from_nexsci._get_nexsci_data()
    assert isinstance(df, pd.DataFrame)
    df = from_nexsci._fill_data(df)
    assert isinstance(df, pd.DataFrame)


def test_lazy_nexsci_local():
    """ Tests if creating a system from the local csv file works """

    # Pass a nice name
    sys = from_nexsci("Kepler-10")
    assert isinstance(sys, starry.System)
    assert len(sys.secondaries) == 2

    # Pass a planet name
    sys = from_nexsci("Kepler-10b")
    assert isinstance(sys, starry.System)
    assert len(sys.secondaries) == 1

    # Pass a bad name
    sys = from_nexsci("kepler 10")
    assert isinstance(sys, starry.System)

    # Pass a random name
    sys = from_nexsci("random")
    assert isinstance(sys, starry.System)



# -*- coding: utf-8 -*-
import logging

rootLogger = logging.getLogger("starry")
rootLogger.addHandler(logging.StreamHandler())
rootLogger.setLevel(logging.INFO)


class ConfigType(type):
    """Global config container."""

    @property
    def rootLogger(cls):
        return rootLogger

    @property
    def rootHandler(cls):
        return rootLogger.handlers[0]

    @property
    def lazy(cls):
        """Indicates whether or not the map evaluates things lazily.

        If True, all attributes and method return values are unevaluated
        ``theano`` nodes. This is particularly useful for model building and
        integration with ``pymc3``. In lazy mode, call the ``.eval()`` method
        on any ``theano`` node to compute and return its numerical value.

        If False, ``starry`` will automatically compile methods called by the
        user, and all methods will return numerical values as in the previous
        version of the code.
        """
        return cls._lazy

    @property
    def quiet(cls):
        """Indicates whether or not to suppress informational messages."""
        return cls._quiet

    @property
    def mode(cls):
        """Enable function profiling in lazy mode."""
        return cls._mode

    @property
    def profile(cls):
        """Enable function profiling in lazy mode."""
        return cls._profile

    @quiet.setter
    def quiet(cls, value):
        cls._quiet = value
        if cls._quiet:
            cls.rootLogger.setLevel(logging.ERROR)
        else:
            cls.rootLogger.setLevel(logging.INFO)

    @lazy.setter
    def lazy(cls, value):
        if (cls._allow_changes) or (cls._lazy == value):
            cls._lazy = value
        else:
            raise Exception(
                "Cannot change the `starry` config at this time. "
                "Config options should be set before instantiating any `starry` maps."
            )

    @profile.setter
    def profile(cls, value):
        if (cls._allow_changes) or (cls._profile == value):
            cls._profile = value
        else:
            raise Exception(
                "Cannot change the `starry` config at this time. "
                "Config options should be set before instantiating any `starry` maps."
            )

    @mode.setter
    def mode(cls, value):
        if (cls._allow_changes) or (cls._mode == value):
            cls._mode = value
        else:
            raise Exception(
                "Cannot change the `starry` config at this time. "
                "Config options should be set before instantiating any `starry` maps."
            )

    def freeze(cls):
        cls._allow_changes = False


class config(metaclass=ConfigType):
    _allow_changes = True
    _lazy = True
    _quiet = False
    _profile = False
    _mode = None



# -*- coding: utf-8 -*-
"""Test map visualization.

These tests don't check for anything; we're just ensuring
the code runs without raising errors.

"""
import matplotlib.pyplot as plt
import starry
import numpy as np
import os
import pytest


# TODO: Several tests segfault on CI. Investigate.
STARRY_TEST_MP4 = False
STARRY_ON_CI = os.getenv("CI", "false") == "true"


def test_show(mp4=False):
    map = starry.Map(ydeg=1, udeg=1)
    map.show(file="tmp.pdf", projection="ortho")
    os.remove("tmp.pdf")
    map.show(file="tmp.pdf", projection="rect")
    os.remove("tmp.pdf")
    if STARRY_TEST_MP4:
        map.show(theta=np.linspace(0, 360, 10), file="tmp.mp4")
        os.remove("tmp.mp4")


def test_show_with_figure():
    map = starry.Map(ydeg=1, udeg=1)
    fig, ax = plt.subplots(1)
    map.show(ax=ax, file="tmp.pdf", projection="ortho")
    os.remove("tmp.pdf")


def test_show_moll():
    map = starry.Map(ydeg=1, udeg=1)
    map.show(file="tmp.pdf", projection="moll")
    os.remove("tmp.pdf")


def test_show_colorbar():
    map = starry.Map(ydeg=1, udeg=1)
    map.show(file="tmp.pdf", projection="ortho", colorbar=True)
    os.remove("tmp.pdf")


def test_show_reflected():
    map = starry.Map(ydeg=1, udeg=1, reflected=True)
    map.show(file="tmp.pdf", projection="ortho")
    os.remove("tmp.pdf")
    map.show(file="tmp.pdf", projection="rect")
    os.remove("tmp.pdf")
    if STARRY_TEST_MP4:
        map.show(theta=np.linspace(0, 360, 10), file="tmp.mp4")
        os.remove("tmp.mp4")


def test_show_rv():
    map = starry.Map(ydeg=1, udeg=1, rv=True)
    map.show(rv=True, file="tmp.pdf", projection="ortho")
    os.remove("tmp.pdf")
    map.show(rv=True, file="tmp.pdf", projection="rect")
    os.remove("tmp.pdf")
    if STARRY_TEST_MP4:
        map.show(rv=True, theta=np.linspace(0, 360, 10), file="tmp.mp4")
        os.remove("tmp.mp4")


def test_show_ld():
    map = starry.Map(udeg=2)
    map.show(file="tmp.pdf")
    os.remove("tmp.pdf")


def test_system_show():
    pri = starry.Primary(starry.Map())
    sec = starry.Secondary(starry.Map(), porb=1.0)
    sys = starry.System(pri, sec)
    sys.show(0.1, file="tmp.pdf")
    os.remove("tmp.pdf")
    if STARRY_TEST_MP4:
        sys.show([0.1, 0.2], file="tmp.mp4")
        os.remove("tmp.mp4")
    sys.show([0.1, 0.2], file="tmp.gif")
    os.remove("tmp.gif")


@pytest.mark.skipif(STARRY_ON_CI, reason="Segfaults on CI")
def test_system_rv_show():
    pri = starry.Primary(starry.Map(rv=True))
    sec = starry.Secondary(starry.Map(rv=True), porb=1.0)
    sys = starry.System(pri, sec)
    sys.show(0.1, file="tmp.pdf")
    os.remove("tmp.pdf")
    if STARRY_TEST_MP4:
        sys.show([0.1, 0.2], file="tmp.mp4")
        os.remove("tmp.mp4")



# -*- coding: utf-8 -*-
"""
Test the various Theano Ops and their gradients.

"""
from starry.compat import theano
from starry.compat import change_flags
from starry.compat import tt
import numpy as np
import pytest
import starry


def test_sT(abs_tol=1e-5, rel_tol=1e-5, eps=1e-7):
    with change_flags(compute_test_value="off"):
        map = starry.Map(ydeg=2)
        theano.gradient.verify_grad(
            map.ops.sT,
            (np.linspace(0.01, 1.09, 30), 0.1),
            abs_tol=abs_tol,
            rel_tol=rel_tol,
            eps=eps,
            n_tests=1,
            rng=np.random,
        )


def test_intensity(abs_tol=1e-5, rel_tol=1e-5, eps=1e-7):
    with change_flags(compute_test_value="off"):
        map = starry.Map(ydeg=2, udeg=2)
        np.random.seed(11)
        lat = 180 * (np.random.random(10) - 0.5)
        lon = 360 * (np.random.random(10) - 0.5)
        y = [1.0] + list(np.random.randn(8))
        u = [-1.0] + list(np.random.randn(2))
        f = [np.pi]
        theta = 0.0

        def intensity(lat, lon, y, u, f, theta):
            return map.ops.intensity(lat, lon, y, u, f, theta, np.array(True))

        theano.gradient.verify_grad(
            intensity,
            (lat, lon, y, u, f, theta),
            abs_tol=abs_tol,
            rel_tol=rel_tol,
            eps=eps,
            n_tests=1,
            rng=np.random,
        )


def test_tensordotRz(abs_tol=1e-5, rel_tol=1e-5, eps=1e-7):
    with change_flags(compute_test_value="off"):
        map = starry.Map(ydeg=2)
        theta = (
            np.array([0.0, 15.0, 30.0, 45.0, 60.0, 75.0, 90.0]) * np.pi / 180.0
        )

        # Matrix M
        M = np.ones((7, 9))
        theano.gradient.verify_grad(
            map.ops.tensordotRz,
            (M, theta),
            abs_tol=abs_tol,
            rel_tol=rel_tol,
            eps=eps,
            n_tests=1,
            rng=np.random,
        )

        # Vector M
        M = np.ones((1, 9))
        theano.gradient.verify_grad(
            map.ops.tensordotRz,
            (M, theta),
            abs_tol=abs_tol,
            rel_tol=rel_tol,
            eps=eps,
            n_tests=1,
            rng=np.random,
        )


def test_dotR(abs_tol=1e-5, rel_tol=1e-5, eps=1e-7):
    with change_flags(compute_test_value="off"):
        map = starry.Map(ydeg=2)

        x = 1.0 / np.sqrt(3)
        y = 1.0 / np.sqrt(3)
        z = 1.0 / np.sqrt(3)
        theta = np.pi / 5

        # Matrix M
        M = np.ones((7, 9))
        theano.gradient.verify_grad(
            map.ops.dotR,
            (M, x, y, z, theta),
            abs_tol=abs_tol,
            rel_tol=rel_tol,
            eps=eps,
            n_tests=1,
            rng=np.random,
        )

        # Vector M
        M = np.ones((1, 9))
        theano.gradient.verify_grad(
            map.ops.dotR,
            (M, x, y, z, theta),
            abs_tol=abs_tol,
            rel_tol=rel_tol,
            eps=eps,
            n_tests=1,
            rng=np.random,
        )


def test_F(abs_tol=1e-5, rel_tol=1e-5, eps=1e-7):
    with change_flags(compute_test_value="off"):
        map = starry.Map(ydeg=2, udeg=2, rv=True)
        np.random.seed(11)
        u = np.random.randn(3)
        u[0] = -1
        f = np.random.randn(16)
        theano.gradient.verify_grad(
            map.ops.F,
            (u, f),
            abs_tol=abs_tol,
            rel_tol=rel_tol,
            eps=eps,
            n_tests=1,
            rng=np.random,
        )


def test_pT(abs_tol=1e-5, rel_tol=1e-5, eps=1e-7):
    with change_flags(compute_test_value="off"):
        map = starry.Map(ydeg=2)
        map[1:, :] = 1
        x = np.array([0.13])
        y = np.array([0.25])
        z = np.sqrt(1 - x ** 2 - y ** 2)
        theano.gradient.verify_grad(
            map.ops.pT,
            (x, y, z),
            abs_tol=abs_tol,
            rel_tol=rel_tol,
            eps=eps,
            n_tests=1,
            rng=np.random,
        )


def test_flux(abs_tol=1e-5, rel_tol=1e-5, eps=1e-7):
    with change_flags(compute_test_value="off"):
        map = starry.Map(ydeg=2)
        theta = np.linspace(0, 30, 10)
        xo = np.linspace(-1.5, 1.5, len(theta))
        yo = np.ones_like(xo) * 0.3
        zo = 1.0 * np.ones_like(xo)
        ro = 0.1
        inc = 85.0 * np.pi / 180.0
        obl = 30.0 * np.pi / 180.0
        y = np.ones(9)
        u = [-1.0]
        f = [np.pi]

        func = lambda *args: tt.dot(map.ops.X(*args), y)

        # Just rotation
        theano.gradient.verify_grad(
            func,
            (theta, xo, yo, zo, 0.0, inc, obl, u, f),
            abs_tol=abs_tol,
            rel_tol=rel_tol,
            eps=eps,
            n_tests=1,
            rng=np.random,
        )

        # Just occultation
        theano.gradient.verify_grad(
            func,
            (theta, xo / 3, yo, zo, ro, inc, obl, u, f),
            abs_tol=abs_tol,
            rel_tol=rel_tol,
            eps=eps,
            n_tests=1,
            rng=np.random,
        )

        # Rotation + occultation
        theano.gradient.verify_grad(
            func,
            (theta, xo, yo, zo, ro, inc, obl, u, f),
            abs_tol=abs_tol,
            rel_tol=rel_tol,
            eps=eps,
            n_tests=1,
            rng=np.random,
        )


def test_rT_reflected(abs_tol=1e-5, rel_tol=1e-5, eps=1e-7):
    with change_flags(compute_test_value="off"):
        map = starry.Map(ydeg=2, reflected=True)
        bterm = np.linspace(-1, 1, 10)[1:-1]
        sigr = 30 * np.pi / 180
        theano.gradient.verify_grad(
            map.ops.rT,
            (bterm, sigr),
            abs_tol=abs_tol,
            rel_tol=rel_tol,
            eps=eps,
            n_tests=1,
            rng=np.random,
        )


"""
# TODO: Implement the gradient of the OrenNayarOp.
def test_intensity_reflected(abs_tol=1e-5, rel_tol=1e-5, eps=1e-7):
    with change_flags(compute_test_value="off"):
        map = starry.Map(ydeg=2, udeg=2, reflected=True)
        np.random.seed(11)
        lat = 180 * (np.random.random(10) - 0.5)
        lon = 360 * (np.random.random(10) - 0.5)
        y = [1.0] + list(np.random.randn(8))
        u = [-1.0] + list(np.random.randn(2))
        f = [np.pi]
        source = np.random.randn(10, 3)
        source /= np.sqrt(np.sum(source ** 2, axis=1)).reshape(-1, 1)
        xs = source[:, 0]
        ys = source[:, 1]
        zs = source[:, 2]
        Rs = 1.0
        theta = 0.0
        sigr = 30 * np.pi / 180

        def intensity(lat, lon, y, u, f, xs, ys, zs, Rs, theta, sigr):
            return map.ops.intensity(
                lat,
                lon,
                y,
                u,
                f,
                xs,
                ys,
                zs,
                Rs,
                theta,
                np.array(False),
                sigr,
                np.array(False),
                np.array(True),
            )

        theano.gradient.verify_grad(
            intensity,
            (lat, lon, y, u, f, xs, ys, zs, Rs, theta, sigr),
            abs_tol=abs_tol,
            rel_tol=rel_tol,
            eps=eps,
            n_tests=1, rng=np.random,
        )
"""


def test_flux_reflected(abs_tol=1e-5, rel_tol=1e-5, eps=1e-7):
    with change_flags(compute_test_value="off"):
        map = starry.Map(ydeg=2, reflected=True)
        theta = np.linspace(0, 30, 10)
        xs = np.linspace(-1.5, 1.5, len(theta))
        ys = np.ones_like(xs) * 0.3
        zs = -1.0 * np.ones_like(xs)
        ro = 0.0
        inc = 85.0 * np.pi / 180.0
        obl = 30.0 * np.pi / 180.0
        y = np.ones(9)
        u = [-1.0]
        f = [np.pi]
        Rs = 1.0
        sigr = 30 * np.pi / 180

        def func(theta, xs, ys, zs, Rs, ro, inc, obl, u, f):
            return tt.dot(
                map.ops.X(
                    theta, xs, ys, zs, Rs, xs, ys, zs, ro, inc, obl, u, f, sigr
                ),
                y,
            )

        # Just rotation
        theano.gradient.verify_grad(
            func,
            (theta, xs, ys, zs, Rs, ro, inc, obl, u, f),
            abs_tol=abs_tol,
            rel_tol=rel_tol,
            eps=eps,
            n_tests=1,
            rng=np.random,
        )


def test_flux_ylm_ld(abs_tol=1e-5, rel_tol=1e-5, eps=1e-7):
    with change_flags(compute_test_value="off"):
        map = starry.Map(ydeg=2, udeg=2)
        theta = np.linspace(0, 30, 10)
        xo = np.linspace(-1.5, 1.5, len(theta))
        yo = np.ones_like(xo) * 0.3
        zo = 1.0 * np.ones_like(xo)
        ro = 0.1
        inc = 85.0 * np.pi / 180.0
        obl = 30.0 * np.pi / 180.0
        y = np.ones(9)
        np.random.seed(14)
        u = [-1.0] + list(np.random.randn(2))
        f = [np.pi]

        func = lambda *args: tt.dot(map.ops.X(*args), y)

        # Just rotation
        theano.gradient.verify_grad(
            func,
            (theta, xo, yo, zo, 0.0, inc, obl, u, f),
            abs_tol=abs_tol,
            rel_tol=rel_tol,
            eps=eps,
            n_tests=1,
            rng=np.random,
        )

        # Just occultation
        theano.gradient.verify_grad(
            func,
            (theta, xo / 3, yo, zo, ro, inc, obl, u, f),
            abs_tol=abs_tol,
            rel_tol=rel_tol,
            eps=eps,
            n_tests=1,
            rng=np.random,
        )

        # Rotation + occultation
        theano.gradient.verify_grad(
            func,
            (theta, xo, yo, zo, ro, inc, obl, u, f),
            abs_tol=abs_tol,
            rel_tol=rel_tol,
            eps=eps,
            n_tests=1,
            rng=np.random,
        )


def test_flux_quad_ld(abs_tol=1e-5, rel_tol=1e-5, eps=1e-7):
    with change_flags(compute_test_value="off"):
        map = starry.Map(udeg=2)
        xo = np.linspace(-1.5, 1.5, 10)
        yo = np.ones_like(xo) * 0.3
        zo = 1.0 * np.ones_like(xo)
        ro = 0.1
        np.random.seed(14)
        u = np.array([-1.0] + list(np.random.randn(2)))
        func = lambda *args: map.ops.flux(*args)

        theano.gradient.verify_grad(
            func,
            (xo, yo, zo, ro, u),
            abs_tol=abs_tol,
            rel_tol=rel_tol,
            eps=eps,
            n_tests=1,
            rng=np.random,
        )


def test_rv(abs_tol=1e-5, rel_tol=1e-5, eps=1e-7):
    with change_flags(compute_test_value="off"):
        map = starry.Map(ydeg=2, rv=True)
        theta = np.linspace(0, 30, 10)
        xo = np.linspace(-1.5, 1.5, len(theta))
        yo = np.ones_like(xo) * 0.3
        zo = 1.0 * np.ones_like(xo)
        ro = 0.1
        inc = 85.0 * np.pi / 180.0
        obl = 30.0 * np.pi / 180.0
        veq = 0.5
        alpha = 0.3
        y = np.ones(9)
        u = [-1.0]

        # Just rotation
        theano.gradient.verify_grad(
            map.ops.rv,
            (theta, xo, yo, zo, 0.0, inc, obl, y, u, veq, alpha),
            abs_tol=abs_tol,
            rel_tol=rel_tol,
            eps=eps,
            n_tests=1,
            rng=np.random,
        )

        # Just occultation
        theano.gradient.verify_grad(
            map.ops.rv,
            (theta, xo / 3, yo, zo, ro, inc, obl, y, u, veq, alpha),
            abs_tol=abs_tol,
            rel_tol=rel_tol,
            eps=eps,
            n_tests=1,
            rng=np.random,
        )

        # Rotation + occultation
        theano.gradient.verify_grad(
            map.ops.rv,
            (theta, xo, yo, zo, ro, inc, obl, y, u, veq, alpha),
            abs_tol=abs_tol,
            rel_tol=rel_tol,
            eps=eps,
            n_tests=1,
            rng=np.random,
        )


def test_spot(abs_tol=1e-5, rel_tol=1e-5, eps=1e-7):
    with change_flags(compute_test_value="off"):
        map = starry.Map(ydeg=5)
        amp = [-0.01]
        sigma = 0.1
        lat = 30 * np.pi / 180
        lon = 45 * np.pi / 180
        theano.gradient.verify_grad(
            map.ops.spotYlm,
            (amp, sigma, lat, lon),
            abs_tol=abs_tol,
            rel_tol=rel_tol,
            eps=eps,
            n_tests=1,
            rng=np.random,
        )


def test_spot_spectral(abs_tol=1e-5, rel_tol=1e-5, eps=1e-7):
    with change_flags(compute_test_value="off"):
        map = starry.Map(ydeg=5, nw=2)
        amp = [-0.01, -0.02]
        sigma = 0.1
        lat = 30 * np.pi / 180
        lon = 45 * np.pi / 180
        theano.gradient.verify_grad(
            map.ops.spotYlm,
            (amp, sigma, lat, lon),
            abs_tol=abs_tol,
            rel_tol=rel_tol,
            eps=eps,
            n_tests=1,
            rng=np.random,
        )


def test_sT_reflected(abs_tol=1e-5, rel_tol=1e-5, eps=1e-7):
    with change_flags(compute_test_value="off"):
        map = starry.Map(ydeg=2, reflected=True)
        b = np.array([0.5])
        theta = np.array([0.5])
        bo = np.array([0.75])
        ro = 0.5
        sigr = 30 * np.pi / 180
        theano.gradient.verify_grad(
            map.ops.sT,
            (b, theta, bo, ro, sigr),
            abs_tol=abs_tol,
            rel_tol=rel_tol,
            eps=eps,
            n_tests=1,
            rng=np.random,
        )



# -*- coding: utf-8 -*-
from ._constants import *
import numpy as np


__all__ = [
    "get_ortho_latitude_lines",
    "get_ortho_longitude_lines",
    "get_moll_latitude_lines",
    "get_moll_longitude_lines",
    "get_projection",
]


def RAxisAngle(axis=[0, 1, 0], theta=0):
    """

    """
    cost = np.cos(theta)
    sint = np.sin(theta)

    return np.reshape(
        [
            cost + axis[0] * axis[0] * (1 - cost),
            axis[0] * axis[1] * (1 - cost) - axis[2] * sint,
            axis[0] * axis[2] * (1 - cost) + axis[1] * sint,
            axis[1] * axis[0] * (1 - cost) + axis[2] * sint,
            cost + axis[1] * axis[1] * (1 - cost),
            axis[1] * axis[2] * (1 - cost) - axis[0] * sint,
            axis[2] * axis[0] * (1 - cost) - axis[1] * sint,
            axis[2] * axis[1] * (1 - cost) + axis[0] * sint,
            cost + axis[2] * axis[2] * (1 - cost),
        ],
        [3, 3],
    )


def get_moll_latitude_lines(dlat=np.pi / 6, npts=1000, niter=100):
    res = []
    latlines = np.arange(-np.pi / 2, np.pi / 2, dlat)[1:]
    for lat in latlines:
        theta = lat
        for n in range(niter):
            theta -= (2 * theta + np.sin(2 * theta) - np.pi * np.sin(lat)) / (
                2 + 2 * np.cos(2 * theta)
            )
        x = np.linspace(-2 * np.sqrt(2), 2 * np.sqrt(2), npts)
        y = np.ones(npts) * np.sqrt(2) * np.sin(theta)
        a = np.sqrt(2)
        b = 2 * np.sqrt(2)
        y[(y / a) ** 2 + (x / b) ** 2 > 1] = np.nan
        res.append((x, y))
    return res


def get_moll_longitude_lines(dlon=np.pi / 6, npts=1000, niter=100):
    res = []
    lonlines = np.arange(-np.pi, np.pi, dlon)[1:]
    for lon in lonlines:
        lat = np.linspace(-np.pi / 2, np.pi / 2, npts)
        theta = np.array(lat)
        for n in range(niter):
            theta -= (2 * theta + np.sin(2 * theta) - np.pi * np.sin(lat)) / (
                2 + 2 * np.cos(2 * theta)
            )
        x = 2 * np.sqrt(2) / np.pi * lon * np.cos(theta)
        y = np.sqrt(2) * np.sin(theta)
        res.append((x, y))
    return res


def get_ortho_latitude_lines(inc=np.pi / 2, obl=0, dlat=np.pi / 6, npts=1000):
    """

    """
    # Angular quantities
    ci = np.cos(inc)
    si = np.sin(inc)
    co = np.cos(obl)
    so = np.sin(obl)

    # Latitude lines
    res = []
    latlines = np.arange(-np.pi / 2, np.pi / 2, dlat)[1:]
    for lat in latlines:

        # Figure out the equation of the ellipse
        y0 = np.sin(lat) * si
        a = np.cos(lat)
        b = a * ci
        x = np.linspace(-a, a, npts)
        y1 = y0 - b * np.sqrt(1 - (x / a) ** 2)
        y2 = y0 + b * np.sqrt(1 - (x / a) ** 2)

        # Mask lines on the backside
        if si != 0:
            if inc > np.pi / 2:
                ymax = y1[np.argmax(x ** 2 + y1 ** 2)]
                y1[y1 < ymax] = np.nan
                ymax = y2[np.argmax(x ** 2 + y2 ** 2)]
                y2[y2 < ymax] = np.nan
            else:
                ymax = y1[np.argmax(x ** 2 + y1 ** 2)]
                y1[y1 > ymax] = np.nan
                ymax = y2[np.argmax(x ** 2 + y2 ** 2)]
                y2[y2 > ymax] = np.nan

        # Rotate them
        for y in (y1, y2):
            xr = -x * co - y * so
            yr = -x * so + y * co
            res.append((xr, yr))

    return res


def get_ortho_longitude_lines(
    inc=np.pi / 2, obl=0, theta=0, dlon=np.pi / 6, npts=1000
):
    """

    """

    # Angular quantities
    ci = np.cos(inc)
    si = np.sin(inc)
    co = np.cos(obl)
    so = np.sin(obl)

    # Are we (essentially) equator-on?
    equator_on = (inc > 88 * np.pi / 180) and (inc < 92 * np.pi / 180)

    # Longitude grid lines
    res = []
    if equator_on:
        offsets = np.arange(-np.pi / 2, np.pi / 2, dlon)
    else:
        offsets = np.arange(0, 2 * np.pi, dlon)

    for offset in offsets:

        # Super hacky, sorry. This can probably
        # be coded up more intelligently.
        if equator_on:
            sgns = [1]
            if np.cos(theta + offset) >= 0:
                bsgn = 1
            else:
                bsgn = -1
        else:
            bsgn = 1
            if np.cos(theta + offset) >= 0:
                sgns = np.array([1, -1])
            else:
                sgns = np.array([-1, 1])

        for lon, sgn in zip([0, np.pi], sgns):

            # Viewed at i = 90
            y = np.linspace(-1, 1, npts)
            b = bsgn * np.sin(lon - theta - offset)
            x = b * np.sqrt(1 - y ** 2)
            z = sgn * np.sqrt(np.abs(1 - x ** 2 - y ** 2))

            if equator_on:

                pass

            else:

                # Rotate by the inclination
                R = RAxisAngle([1, 0, 0], np.pi / 2 - inc)
                v = np.vstack(
                    (x.reshape(1, -1), y.reshape(1, -1), z.reshape(1, -1))
                )
                x, y, _ = np.dot(R, v)

                # Mask lines on the backside
                if si != 0:
                    if inc < np.pi / 2:
                        imax = np.argmax(x ** 2 + y ** 2)
                        y[: imax + 1] = np.nan
                    else:
                        imax = np.argmax(x ** 2 + y ** 2)
                        y[imax:] = np.nan

            # Rotate by the obliquity
            xr = -x * co - y * so
            yr = -x * so + y * co
            res.append((xr, yr))

    return res


def get_projection(projection):
    """

    """
    if projection.lower().startswith("rect"):
        projection = STARRY_RECTANGULAR_PROJECTION
    elif projection.lower().startswith("ortho"):
        projection = STARRY_ORTHOGRAPHIC_PROJECTION
    elif projection.lower().startswith("moll"):
        projection = STARRY_MOLLWEIDE_PROJECTION
    else:
        raise ValueError("Unknown map projection.")
    return projection



# -*- coding: utf-8 -*-
from ....compat import Apply, tt
from .base_op import LimbDarkBaseOp


__all__ = ["GetClRevOp"]


class GetClRevOp(LimbDarkBaseOp):

    __props__ = ()
    func_file = "./get_cl_rev.cc"
    func_name = "APPLY_SPECIFIC(get_cl_rev)"

    def make_node(self, bc):
        return Apply(self, [tt.as_tensor_variable(bc)], [bc.type()])

    def infer_shape(self, *args):
        shapes = args[-1]
        return (shapes[0],)



# -*- coding: utf-8 -*-
from ...compat import Apply, Op, tt, ts
import numpy as np
import os

# C extensions are not installed on RTD
if os.getenv("READTHEDOCS") == "True":
    STARRY_OREN_NAYAR_DEG = 5
else:
    from ..._c_ops import STARRY_OREN_NAYAR_DEG


__all__ = ["FOp", "OrenNayarOp"]


class FOp(Op):
    def __init__(self, func, N, Ny):
        self.func = func
        self.N = N
        self.Ny = Ny
        self._grad_op = FGradientOp(self)

    def make_node(self, *inputs):
        inputs = [tt.as_tensor_variable(i) for i in inputs]
        outputs = [tt.TensorType(inputs[-1].dtype, (False, False))()]
        return Apply(self, inputs, outputs)

    def infer_shape(self, *args):
        return [(self.N, self.Ny)]

    def R_op(self, inputs, eval_points):
        if eval_points[0] is None:
            return eval_points
        return self.grad(inputs, eval_points)

    def perform(self, node, inputs, outputs):
        outputs[0][0] = self.func(*inputs)

    def grad(self, inputs, gradients):
        return self._grad_op(*(inputs + gradients))


class FGradientOp(Op):
    def __init__(self, base_op):
        self.base_op = base_op

    def make_node(self, *inputs):
        inputs = [tt.as_tensor_variable(i) for i in inputs]
        outputs = [i.type() for i in inputs[:-1]]
        return Apply(self, inputs, outputs)

    def infer_shape(self, *args):
        shapes = args[-1]
        return shapes[:-1]

    def perform(self, node, inputs, outputs):
        bu, bf = self.base_op.func(*inputs)
        outputs[0][0] = np.reshape(bu, np.shape(inputs[0]))
        outputs[1][0] = np.reshape(bf, np.shape(inputs[1]))


class OrenNayarOp(Op):
    def __init__(self, func):
        self.func = func

    def make_node(self, *inputs):
        inputs = [tt.as_tensor_variable(i) for i in inputs]
        outputs = [tt.TensorType(inputs[-1].dtype, (False, False))()]
        return Apply(self, inputs, outputs)

    def infer_shape(self, *args):
        shapes = args[-1]
        return [((STARRY_OREN_NAYAR_DEG + 1) ** 2, shapes[0][0])]

    def R_op(self, inputs, eval_points):
        if eval_points[0] is None:
            return eval_points
        return self.grad(inputs, eval_points)

    def perform(self, node, inputs, outputs):
        outputs[0][0] = self.func(*inputs)



"""Install script for `starry`."""
from setuptools import setup, Extension, find_packages
from setuptools.command.build_ext import build_ext
import warnings
import setuptools
import subprocess
import sys
import os
import glob

# Are we on ReadTheDocs?
on_rtd = os.getenv("READTHEDOCS") == "True"

# Custom compiler flags
macros = dict(
    STARRY_NDIGITS=16,
    STARRY_ELLIP_MAX_ITER=200,
    STARRY_MAX_LMAX=50,
    STARRY_BCUT=1.0e-3,
    STARRY_MN_MAX_ITER=100,
    STARRY_IJ_MAX_ITER=200,
    STARRY_REFINE_J_AT=25,
    STARRY_USE_INCOMPLETE_INTEGRALS=0,
    STARRY_QUAD_POINTS=100,
    STARRY_EL2_MAX_ITER=100,
)

# Override with user values
for key, value in macros.items():
    macros[key] = os.getenv(key, value)

# Compiler optimization flag -O
optimize = int(os.getenv("STARRY_O", 2))
assert optimize in [0, 1, 2, 3], "Invalid optimization flag."
macros["STARRY_O"] = optimize

# Branching optimizations (disable on Windows)
disable_branch_optim = int(os.getenv("STARRY_BRANCHING_DISABLE_OPTIM", 0))
if sys.platform not in ["darwin", "linux"]:
    disable_branch_optim = 1
macros["STARRY_BRANCHING_DISABLE_OPTIM"] = disable_branch_optim

# Debug mode?
debug = bool(int(os.getenv("STARRY_DEBUG", 0)))
if debug:
    optimize = 0
    macros["STARRY_O"] = 0
    macros["STARRY_DEBUG"] = 1

# Numerical override at high l?
if bool(int(os.getenv("STARRY_KL_NUMERICAL", 0))):
    macros["STARRY_KL_NUMERICAL"] = 1

# Compute the Oren-Nayar (1994) expansion if the user requests it
deg = os.getenv("STARRY_OREN_NAYAR_DEG", None)
Nb = os.getenv("STARRY_OREN_NAYAR_NB", None)
if (deg is not None) or (Nb is not None):

    sys.path.insert(
        1,
        os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "starry",
            "_core",
            "ops",
            "lib",
            "include",
            "reflected",
        ),
    )
    import oren_nayar

    kwargs = dict()
    if deg is not None:
        kwargs["deg"] = int(deg)
    if Nb is not None:
        kwargs["Nb"] = int(Nb)
    oren_nayar.generate_header(**kwargs)


class get_pybind_include(object):
    """
    Helper class to determine the pybind11 include path.

    The purpose of this class is to postpone importing pybind11
    until it is actually installed, so that the ``get_include()``
    method can be invoked.
    """

    def __init__(self, user=False):
        self.user = user

    def __str__(self):
        import pybind11

        return pybind11.get_include(self.user)


# As of Python 3.6, CCompiler has a `has_flag` method.
# cf http://bugs.python.org/issue26689
def has_flag(compiler, flagname):
    """Return a boolean indicating whether a flag name is supported on
    the specified compiler.
    """
    import tempfile

    with tempfile.NamedTemporaryFile("w", suffix=".cpp") as f:
        f.write("int main (int argc, char **argv) { return 0; }")
        try:
            compiler.compile([f.name], extra_postargs=["-w", flagname])
        except setuptools.distutils.errors.CompileError:
            return False
    return True


def cpp_flag(compiler):
    """Return the -std=c++[11/14/17] compiler flag.

    The newer version is prefered over c++11 (when it is available).
    """
    flags = ["-std=c++17", "-std=c++14", "-std=c++11"]

    for flag in flags:
        if has_flag(compiler, flag):
            return flag

    raise RuntimeError(
        "Unsupported compiler -- at least C++11 support is needed!"
    )


class BuildExt(build_ext):
    """A custom build extension for adding compiler-specific options."""

    c_opts = {"msvc": ["/EHsc"], "unix": []}
    l_opts = {"msvc": [], "unix": []}

    if sys.platform == "darwin":
        darwin_opts = ["-stdlib=libc++", "-mmacosx-version-min=10.14"]
        c_opts["unix"] += darwin_opts
        l_opts["unix"] += darwin_opts

    def build_extensions(self):
        ct = self.compiler.compiler_type
        opts = self.c_opts.get(ct, [])
        link_opts = self.l_opts.get(ct, [])
        if ct == "unix":
            opts.append(
                '-DVERSION_INFO="%s"' % self.distribution.get_version()
            )
            opts.append(cpp_flag(self.compiler))
            if has_flag(self.compiler, "-fcolor-diagnostics"):
                opts.append("-fcolor-diagnostics")
        elif ct == "msvc":
            opts.append(
                '/DVERSION_INFO=\\"%s\\"' % self.distribution.get_version()
            )
            opts.append("/Zm10")  # debug for C1060
        extra_args = ["-O%d" % optimize]
        if debug:
            extra_args += ["-g", "-Wall", "-fno-lto"]
        else:
            if has_flag(self.compiler, "-fvisibility=hidden"):
                opts.append("-fvisibility=hidden")
            extra_args += ["-g0"]
        for ext in self.extensions:
            ext.extra_compile_args = opts + extra_args
            ext.extra_link_args = link_opts + extra_args

        build_ext.build_extensions(self)


if int(macros["STARRY_NDIGITS"]) > 16:
    include_dirs = ["starry/_core/ops/lib/vendor/boost_1_66_0"]
else:
    include_dirs = []
if on_rtd:
    ext_modules = []
    cmdclass = {}
else:
    ext_modules = [
        Extension(
            "starry._c_ops",
            ["starry/_core/ops/lib/include/interface.cpp"],
            include_dirs=include_dirs
            + [
                # Path to pybind11 headers
                get_pybind_include(),
                get_pybind_include(user=True),
                "starry/_core/ops/lib/include",
                "starry/_core/ops/lib/vendor/eigen_3.3.5",
            ],
            language="c++",
            define_macros=[(key, value) for key, value in macros.items()],
        )
    ]
    cmdclass = {"build_ext": BuildExt}


setup(
    name="starry",
    author="Rodrigo Luger",
    author_email="rodluger@gmail.com",
    url="https://github.com/rodluger/starry",
    description="Analytic occultation light curves for astronomy.",
    long_description=open("README.md", "r").read(),
    long_description_content_type="text/markdown",
    license="MIT",
    packages=find_packages(),
    ext_modules=ext_modules,
    use_scm_version={
        "write_to": os.path.join("starry", "starry_version.py"),
        "write_to_template": '__version__ = "{version}"\n',
    },
    install_requires=[
        "setuptools_scm",
        "numpy>=1.19.2",
        "scipy>=1.5.0",
        "astropy>=3.1",
        "pymc3",
        "pymc3-ext",
        "aesara-theano-fallback>=0.0.4",
        "matplotlib>=3.2.2",
        "ipython",
        "exoplanet>=0.4.5",
        "packaging",
    ],
    extras_require={
        "tests": [
            "parameterized",
            "nose",
            "pytest",
            "pytest-dependency",
            "pytest-env",
            "pytest-cov",
            "scikit-image",
            "pillow",
            "tqdm",
        ],
        "docs": [
            "sphinx>=1.7.5",
            "pandoc",
            "jupyter",
            "jupytext",
            "ipywidgets",
            "nbformat",
            "nbconvert",
            "rtds_action",
            "nbsphinx",
            "tqdm",
        ],
    },
    setup_requires=["setuptools_scm", "pybind11>2.4"],
    cmdclass=cmdclass,
    data_files=glob.glob("starry/img/*.png"),
    include_package_data=True,
    zip_safe=False,
)



# -*- coding: utf-8 -*-
from .exceptions import *
from .filter import *
from .integration import *
from .limbdark import *
from .minimize import *
from .polybasis import *
from .rotation import *
from .spot import *



# -*- coding: utf-8 -*-
from ...compat import Apply, Op, tt, floatX
import numpy as np

__all__ = ["CheckBoundsOp", "RaiseValueErrorOp", "RaiseValueErrorIfOp"]


class CheckBoundsOp(Op):
    """

    """

    def __init__(self, lower=-np.inf, upper=np.inf, name=None):
        self.lower = lower
        self.upper = upper
        if name is None:
            self.name = "parameter"
        else:
            self.name = name

    def make_node(self, *inputs):
        inputs = [tt.as_tensor_variable(inputs[0])]
        outputs = [inputs[0].type()]
        return Apply(self, inputs, outputs)

    def infer_shape(self, *args):
        shapes = args[-1]
        return [shapes[0]]

    def perform(self, node, inputs, outputs):
        outputs[0][0] = inputs[0]
        if np.any((inputs[0] < self.lower) | (inputs[0] > self.upper)):
            low = np.where((inputs[0] < self.lower))[0]
            high = np.where((inputs[0] > self.upper))[0]
            if len(low):
                value = inputs[0][low[0]]
                sign = "<"
                bound = self.lower
            else:
                value = inputs[0][high[0]]
                sign = ">"
                bound = self.upper
            raise ValueError(
                "%s out of bounds: %f %s %f" % (self.name, value, sign, bound)
            )


class RaiseValueErrorIfOp(Op):
    """

    """

    def __init__(self, message=None):
        self.message = message

    def make_node(self, *inputs):
        condition = inputs
        inputs = [tt.as_tensor_variable(condition)]
        outputs = [tt.TensorType(floatX, ())()]
        return gof.Apply(self, inputs, outputs)

    def infer_shape(self, *args):
        return [()]

    def perform(self, node, inputs, outputs):
        outputs[0][0] = np.array(0.0)
        if inputs[0]:
            raise ValueError(self.message)

    def grad(self, inputs, gradients):
        return [inputs[0] * 0.0]


def RaiseValueErrorOp(msg, shape=()):
    return tt.zeros(shape) * RaiseValueErrorIfOp(msg)(True)
